import java.io.*;

/**
 * Tof = Toffoli
 * This class writes English and Picture files,
 * in the Toffoli SEO compilation mode 
 * (an expansion in terms of Toffoli 
 * gates and CNOTs and 1-quibt gates),
 * for a multi-CNOT with
 * 
 * control bits at 0, 1, 2 , ..., numControls-1
 * the target bit at numBits - 1,
 * where numControls = numBits/2 (integer division!)
 * numControls >= 3, numBits >= 6
 * 
 * This class implements the algorithm for
 * expanding multiCNOTs in terms of Toffolis
 * given in the paper
 * Phys.Rev. A 1995(52)3457-3467 by Barenco et al
 * on page 3463
 */

public class MultiCNotTof1SEO_writer  extends SEO_writer{
	boolean[] 		_controlTypes;
	
	/**
	 * 
	 */
	public  MultiCNotTof1SEO_writer(
        String 			filePrefix, 
        int 			numBits,
        boolean[] 		controlTypes,
        GitsMonitor		gm)
	{ 	
	 	
    	super(filePrefix, numBits);
    	
    	Throw.If(_numBits<6 , 
			"multiCNot must have at least 6 qubits " +
			"(3 control, 1 target, 2 workSpace)");

    	Throw.If( controlTypes.length != _numBits/2, 
			"multiCNot must have numControls = _numBits/2");

    	_controlTypes = controlTypes;
    	_gm = gm;
    	
    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public  MultiCNotTof1SEO_writer(
		PrintWriter  	englishOut,
		PrintWriter  	pictureOut,
		Boolean 		doWrite,
		CktEmbedder 	emb,
	    boolean[] 		controlTypes,
	    GitsMonitor		gm)
    {
    	super(englishOut, pictureOut, emb);
    	//numBitsBef = _numBits
 
    	Throw.If(_numBits<6 , 
			"multiCNot must have at least 6 qubits " +
			"(3 control, 1 target, 2 workSpace)");

    	Throw.If( controlTypes.length != _numBits/2, 
			"multiCNot must have numControls = _numBits/2");
		
    	_controlTypes = controlTypes;
    	_gm = gm;


    	if(doWrite){write();}
    }
    /**
	 * 
	 */
	public void writeSmallerPeak(){
		
		int numControls = _numBits/2; //integer division!
		int workSpaceSize = _numBits - 1 - numControls;
		//example:
		//if numBits=9, numControls=4, workSpaceSize=4
		//if numBits=8, numControls=4, workSpaceSize=3

		TFControls  toffoliControls = new TFControls(_numBits, 2);
		TFControls  cNotControls = new TFControls(_numBits, 1);
		OneBitGate xx = new SigX();

		//right side (chain of Toffolis)
		int tarBitPos = _numBits - 2;
		int trolBitPosLow = numControls - 2;
		int trolBitPosHigh = trolBitPosLow + workSpaceSize;
		while(trolBitPosLow>0){
			toffoliControls.setControl(0, trolBitPosHigh, true);
			toffoliControls.setControl(1, trolBitPosLow, true);
			writeControlledOneBitGate(
					tarBitPos,
					toffoliControls,
					xx);
			tarBitPos--;
			trolBitPosLow--;
			trolBitPosHigh--;
		}

		//peak (always a CNOT)
		boolean evenNumBits = (_numBits%2==0?true:false);
		tarBitPos = (evenNumBits?numControls:numControls + 1);
		cNotControls.setControl(0, 0/*bitPos*/, true);
		writeControlledOneBitGate(
			tarBitPos,
			cNotControls,
			xx);

		//left side (chain of Toffolis)
		tarBitPos++;
		trolBitPosLow = 1;
		trolBitPosHigh = trolBitPosLow + workSpaceSize;
		while(tarBitPos< _numBits - 1){
			toffoliControls.setControl(0, trolBitPosHigh, true);
			toffoliControls.setControl(1, trolBitPosLow, true);
			writeControlledOneBitGate(
				tarBitPos,
				toffoliControls,
				xx);
			tarBitPos++;
			trolBitPosLow++;
			trolBitPosHigh++;
		}

	}


	/**
	 * 
	 */
	public void write(){
		int numControls = _numBits/2; //integer division!
		int workSpaceSize = _numBits - 1 - numControls;
		//example:
		//if numBits=9, numControls=4, workSpaceSize=4
		//if numBits=8, numControls=4, workSpaceSize=3

		int j;
		OneBitGate xx = new SigX();
		for(j=0; j<numControls; j++){
			if(_controlTypes[j]==false){
				writeOneBitGate(j, xx);
			}
		}
		

		TFControls  toffoliControls = new TFControls(_numBits, 2);
		int tarBitPos = _numBits - 1;
		int trolBitPosHigh = tarBitPos - 1;
		int trolBitPosLow = trolBitPosHigh - workSpaceSize;
		toffoliControls.setControl(0, trolBitPosHigh, true);
		toffoliControls.setControl(1, trolBitPosLow, true);

		writeControlledOneBitGate(
			tarBitPos,
			toffoliControls,
			xx);
		writeSmallerPeak();
		writeControlledOneBitGate(
			tarBitPos,
			toffoliControls,
			xx);	
		writeSmallerPeak();		
		
		for(j=0; j<numControls; j++){
			if(_controlTypes[j]==false){
				writeOneBitGate(j, xx);
			}
		}

	}
	/**
	 * 
	 */ 
   public static void main(String[] args){
	   String	filePrefix = "tof1";
	   int numBits = 8;
	   //checked 6,7,8
	   
	   int numControls = numBits/2;
//	   boolean[] controlTypes = new boolean[numControls];//default false
//	   boolean[] controlTypes = Globals.T_vec(numControls);
	   boolean[] controlTypes = Globals.F_vec(numControls);
	   controlTypes[0] = true;
	   controlTypes[1] = true;
       
       new MultiCNotTof1SEO_writer(
           filePrefix,
           numBits,
           controlTypes,
           new GitsMonitor(numBits, 0));
          
	  	
	  	MultiCNotUinMat   Uin = 
	  		new MultiCNotUinMat(numBits, controlTypes);
//	  	System.out.println(Uin);
	  	
    	SEO_readerMu	reader = new SEO_readerMu(
    	    filePrefix,
    	    numBits, 
    	    false); // useSparse
    	int lenOfSEO = reader.getLenOfSEO();
 //   	System.out.println(reader);
    	
    	double error = 0.0;
    	int row, col;
    	int numRows = 1<<numBits;
    	for(row=0; row<numRows; row++){
    		for(col=0; col<numRows; col++){
    			error += Complex.abs(Complex.minus(
    					Uin.getEntry(row,col), 
    					reader.getEntry(row,col)
    					));
    		}
    	}
    	
    	System.out.println("Number of Bits. = " + numBits);
        System.out.println("Number of Elem. Ops. = " + lenOfSEO);
        System.out.printf("Error = %e\n", error);
   	
	}
}
