import java.io.*;

/**
 * Tof = Toffoli
 * This class writes English and Picture files,
 * in the Toffoli SEO compilation mode 
 * (an expansion in terms of Toffoli 
 * gates and CNOTs and 1-quibt gates),
 * for a multi-CNOT with
 * 
 * control bits at 0, 1, 2 , ..., numControls-1
 * the target bit at numBits - 1,
 * where numControls = numBits-2
 * (hence 1 target, 1 workspace and rest controls)
 * numControls >= 3, numBits >= 5
 * 
 * This class implements the algorithm for
 * expanding multiCNOTs in terms of Toffolis
 * given in the paper
 * Phys.Rev. A 1995(52)3457-3467 by Barenco et al
 * on page 3463
 */

public class MultiCNotTof2SEO_writer  extends SEO_writer{
	boolean[] 		_controlTypes;
	
	/**
	 * 
	 */
	public  MultiCNotTof2SEO_writer(
        String 			filePrefix, 
        int 			numBits,
        boolean[] 		controlTypes,
        GitsMonitor		gm)
	{ 	
	 	
    	super(filePrefix, numBits);
    	
    	Throw.If(_numBits<5 , 
			"multiCNot must have at least 5 qubits " +
			"(3 control, 1 target, 1 workSpace)");

    	Throw.If( controlTypes.length != _numBits - 2, 
			"multiCNot must have numControls = _numBits - 2");

    	_controlTypes = controlTypes;
    	_gm = gm;
    	
    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public  MultiCNotTof2SEO_writer(
		PrintWriter  	englishOut,
		PrintWriter  	pictureOut,
		Boolean 		doWrite,
		CktEmbedder 	emb,
	    boolean[] 		controlTypes,
	    GitsMonitor		gm)
    {
    	super(englishOut, pictureOut, emb);
    	//numBitsBef = _numBits
 
    	Throw.If(_numBits<5 , 
			"multiCNot must have at least 5 qubits " +
			"(3 control, 1 target, 1 workSpace)");

    	Throw.If( controlTypes.length != _numBits - 2, 
			"multiCNot must have numControls = _numBits - 2");
		
    	_controlTypes = controlTypes;
    	_gm = gm;


    	if(doWrite){write();}
    }
	/**
	 * 
	 */
	public void writeSqrt(){
		
		int numBitsBef, numBitsAft, j;
		int[] bitMap;
		boolean[] isP0Control, isP1Control;
		CktEmbedder	preEmb;
		TFControls  toffoliControls = new TFControls(_numBits, 2);

		if((_numBits-1)/2==2){//2 controls			
			toffoliControls.setControl(0, 1/*bitPos*/, true);
			toffoliControls.setControl(1, 0/*bitPos*/, true);
			writeControlledOneBitGate(
				_numBits-2,//targetbitPos
				toffoliControls,
				new SigX());
		}else{//more than 2 controls
			//parameters for preEmb;
			numBitsBef = _numBits-1;
			numBitsAft = _numBits;
			bitMap = new  int[numBitsBef];
			for(j=0; j<numBitsBef; j++){
				bitMap[j] = j;
			}
			isP0Control = Globals.F_vec(numBitsAft);
			isP1Control = Globals.F_vec(numBitsAft);
			preEmb = new  CktEmbedder(
				numBitsBef,
				numBitsAft,
				bitMap,
				isP0Control,
				isP1Control );
			new MultiCNotTof1SEO_writer(
				_englishOut,
				_pictureOut,
				true,//doWrite,
				preEmb.getCompositeMapOfYouFollowedBy(_emb),
				Globals.T_vec(numBitsBef/2), //_controlTypes
				_gm
			);
		}

		if(_numBits/2==2){//2 controls
			toffoliControls.setControl(0, _numBits-2/*bitPos*/, true);
			toffoliControls.setControl(1, _numBits-3/*bitPos*/, true);
			writeControlledOneBitGate(
				_numBits-1,//targetbitPos
				toffoliControls,
				new SigX());			
		}else{//more than 2 controls
			//parameters for preEmb;
			numBitsBef = _numBits;
			numBitsAft = _numBits;
			bitMap = new  int[numBitsBef];
			for(j=0; j<_numBits-1; j++){
				bitMap[j] = _numBits - 2 - j;
			}
			bitMap[_numBits-1] = _numBits-1;
			isP0Control = Globals.F_vec(numBitsAft);
			isP1Control = Globals.F_vec(numBitsAft);
			preEmb = new  CktEmbedder(
				numBitsBef,
				numBitsAft,
				bitMap,
				isP0Control,
				isP1Control );
			new MultiCNotTof1SEO_writer(
				_englishOut,
				_pictureOut,
				true,//doWrite,
				preEmb.getCompositeMapOfYouFollowedBy(_emb),
				Globals.T_vec(numBitsBef/2), //_controlTypes
				_gm);
		}

		
		
	}
	/**
	 * 
	 */
	public void write(){
		int numControls = _numBits - 2;

		int j;
		OneBitGate xx = new SigX();
		for(j=0; j<numControls; j++){
			if(_controlTypes[j]==false){
				writeOneBitGate(j, xx);
			}
		}
		
		writeSqrt();
		writeSqrt();
		
		for(j=0; j<numControls; j++){
			if(_controlTypes[j]==false){
				writeOneBitGate(j, xx);
			}
		}

	}
	/**
	 * 
	 */ 
	public static void main(String[] args){
		String	filePrefix = "tof2";
		int numBits = 8;
		//checked 5,6,7,8
	   
		int numControls = numBits - 2;
//	   	boolean[] controlTypes = new boolean[numControls];//default false
//	   	boolean[] controlTypes = Globals.T_vec(numControls);
		boolean[] controlTypes = Globals.F_vec(numControls);
		controlTypes[0] = true;
		controlTypes[1] = true;
       
		new MultiCNotTof2SEO_writer(
           filePrefix,
           numBits,
           controlTypes,
           new GitsMonitor(numBits, 0));
          
	  	
		MultiCNotUinMat   Uin = 
	  		new MultiCNotUinMat(numBits, controlTypes);
//	  	System.out.println(Uin);
	  	
		SEO_readerMu	reader = new SEO_readerMu(
    	    filePrefix,
    	    numBits, 
    	    false); // useSparse
    	int lenOfSEO = reader.getLenOfSEO();
 //   	System.out.println(reader);

    	double error = 0.0;
    	int row, col;
    	int numRows = 1<<numBits;
    	for(row=0; row<numRows; row++){
    		for(col=0; col<numRows; col++){
    			error += Complex.abs(Complex.minus(
    					Uin.getEntry(row,col), 
    					reader.getEntry(row,col)
    					));
    		}
    	}
    	
    	System.out.println("Number of Bits. = " + numBits);
        System.out.println("Number of Elem. Ops. = " + lenOfSEO);
        System.out.printf("Error = %e\n", error);
   	
	}
}
