import java.io.*;

/**
 * This class writes English and Picture files,
 * in the Exact SEO compilation mode 
 * (see arxiv:quant-ph/9902062 and arxiv:0706.0479)
 * for a multiplexor of the form
 * [ cc, ss]
 * [-ss, cc]
 * where cc and ss are diagonal matrices satisfying 
 * cc^2 + ss^2=1
 */

public class MultiplexorExactSEO_writer  extends SEO_writer{
    double[]  		_radAngles; //all internal angles in radians
    double[] 		_conjugateRadAngles;

	/**
	 * 
	 */	
	public MultiplexorExactSEO_writer(
	        String 		filePrefix, 
	        int 		numBits,  
	  	    double[]  	radAngles)//all internal angles in radians
	{ 	
	 	
    	super(filePrefix, numBits);
    	
    	Throw.If(_numBits<2 , 
			"multiplexor must have at least 2 qubits");

		
		Throw.If(radAngles.length != 1<<(_numBits-1), 
			"wrong  number of angles");		
		_radAngles = radAngles;

    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public MultiplexorExactSEO_writer(
		PrintWriter  englishOut,
		PrintWriter  pictureOut,
		Boolean 	doWrite,
		CktEmbedder emb,  
        double[] 	radAngles)
    {
    	super(englishOut, pictureOut, emb);
    	//numBitsBef = _numBits
 
    	Throw.If(_numBits<2 , 
		"multiplexor must have at least 2 qubits");
		
		Throw.If(radAngles.length != 1<<(_numBits-1), 
			"wrong  number of angles");		
		_radAngles = radAngles;

    	if(doWrite){write();}
    }
      	
		
	/**
	 * 
	 */
	public void write(boolean wantHermitian)
	{	
		//already checked _numBits>=2
		int	numControls = _numBits - 1;	
		_radAngles = Globals.limitedRadians(_radAngles);
		
		if(Globals.isZero(_radAngles)){
			return;
		}
	
		_conjugateRadAngles = HadamardTransform.of(numControls, _radAngles);

		int		numFactors = (1 << numControls);
		int		f=0;
		int		prevONbit, curONbit;//prev=previous, cur = current
		BitVector	curBitVec = new BitVector(numControls, 0);//start at zero
		BitVector	prevActiveBitVec = new BitVector(numControls, 0);
		BitVector	reducedBitVec = new BitVector(_numBits, 0);
			//one more bit than curBitVec and prevActiveBitVec	
		TFControls  trols = new TFControls(_numBits, 1);
		trols.setType(0, true);

		double normalization = Math.pow(Math.sqrt(2), numControls);
		while(f<numFactors){
			double	ang = 
				_conjugateRadAngles[curBitVec.getDecRep()]/normalization;
			if(Globals.isZero(ang)){
				//do nothing
			}else{
				//an active bvec is one for which ang is NOT zero		
				reducedBitVec.setDecRep( 
						curBitVec.getDecRep() ^ prevActiveBitVec.getDecRep());			
				prevONbit = numControls;
				curONbit = reducedBitVec.findONbitToMyRight(prevONbit);
				while(curONbit!=-1){				
					trols.setBitPos(0, curONbit);
					writeControlledOneBitGate(
				        numControls,//targetBitPosition,
				        trols,
				        new SigX());
					prevONbit = curONbit;
					curONbit = reducedBitVec.findONbitToMyRight(prevONbit);
				}
				writeOneBitGate(
					numControls, //targetBitPosition,
				    new OneBitRot(wantHermitian?-ang:ang, OneBitRot.Y_AXIS));			
				prevActiveBitVec.copy(curBitVec);
			}//zero angles if()	
			f++;	
			curBitVec.lazyAdvance(f);
		}//f loop
		
		//Don't forget the leftmost c-nots:
		//reducedBitVec and prevActiveBitVec don't have
		//the same length so can't just equate them.
		reducedBitVec.setDecRep(prevActiveBitVec.getDecRep());
		prevONbit = numControls;
		curONbit = reducedBitVec.findONbitToMyRight(prevONbit);
		while(curONbit!=-1){
			trols.setBitPos(0, curONbit);
			writeControlledOneBitGate(
		        numControls,//targetBitPosition,
		        trols,
		        new SigX());
			prevONbit = curONbit;
			curONbit = reducedBitVec.findONbitToMyRight(prevONbit);
		}
	
	}
	/**
	 * 
	 */
	public void write(){
		write(false);
	}
	/**
	 * 
	 */
	public void writeHermitian(){
		write(true);
    }

	
	/**
	 * 
	 */ 
	public static void main(String[] args){
		
		String	filePrefix = "exact";
		int numBits = 4;
		
		int numAngles = 1<<(numBits-1);
	  	double[]  radAngles = new  double[numAngles];
	  	for(int j=0; j<numAngles; j++){
	  		radAngles[j] = (Math.random() -.5)*23;
	  	}

	  	new MultiplexorExactSEO_writer(
	  			filePrefix, 
		        numBits,  
		  	    radAngles);
	  	
	  	
	  	MultiplexorUinMat   Uin = new MultiplexorUinMat(numBits, 
	  			new DoubleArray(radAngles));
	  	//System.out.println(Uin);
	  	
    	SEO_readerMu	reader = new SEO_readerMu(
    	    filePrefix,
    	    numBits, 
    	    false); // useSparse
    	int lenOfSEO = reader.getLenOfSEO();
    	//System.out.println(reader);
    	
    	double error = 0.0;
    	int row, col;
    	int numRows = 1<<numBits;
    	for(row=0; row<numRows; row++){
    		for(col=0; col<numRows; col++){
    			error += Complex.abs(Complex.minus(
    					Uin.getEntry(row,col), 
    					reader.getEntry(row,col)
    					));
    		}
    	}
    	
    	System.out.println("Number of Bits. = " + numBits);
        System.out.println("Number of Elem. Ops. = " + lenOfSEO);
        System.out.printf("Error = %e\n", error);
   	
	}
}
