import java.io.*;
/**
 * This class writes English and Picture files,
 * in the Oracular Approximation compilation mode
 * (see arxiv 0901.3851v2)
 * for a multiplexor of the form
 * [ cc, ss]
 * [-ss, cc]
 * where cc and ss are diagonal matrices satisfying 
 * cc^2 + ss^2=1
 */
public class MultiplexorOracularSEO_writer   extends SEO_writer{
 
    int				_numGits;
    int 			_bitPrecision;   
    DoubleArray  	_radAngles; //all internal angles in radians


	/**
	 * 
	 */	
	public MultiplexorOracularSEO_writer(
	        String 		filePrefix, 
	        int 		numBits,  
	  	    DoubleArray radAngles,//all internal angles in radians
	  	    int 		numGits,
	  	    int			bitPrecision)
	{ 	
	 	
    	super(filePrefix, numBits);    	
    	_numGits = numGits;
    	_bitPrecision = bitPrecision;

    	Throw.If(_numBits-_numGits<2 , 
			"multiplexor must have at least 1 multicontrol and 1 target");
    	Throw.If(bitPrecision>numGits, 
			"Insufficiently many grounded bits for requested bit precision");

		int numControls = _numBits - _numGits -1;		

		Throw.If(radAngles.length() != 1<<numControls, 
			"wrong  number of angles");		
		_radAngles = radAngles;

    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public MultiplexorOracularSEO_writer(
		PrintWriter  englishOut,
		PrintWriter  pictureOut,
		Boolean 	doWrite,
		CktEmbedder emb,  
        DoubleArray radAngles,//all internal angles in radians
  	    int 		numGits,
  	    int			bitPrecision)
    {
    	super(englishOut, pictureOut, emb);
    	//numBitsBef = _numBits
 
    	_numGits = numGits;
    	_bitPrecision = bitPrecision;

    	Throw.If(_numBits-_numGits<2 , 
			"multiplexor must have at least 1 multicontrol and 1 target");
    	Throw.If(bitPrecision>numGits, 
			"Insufficiently many grounded bits for requested bit precision");

		int numControls = _numBits - _numGits -1;		

		Throw.If(radAngles.length() != 1<<numControls, 
			"wrong  number of angles");		
		_radAngles = radAngles;

    	if(doWrite){write();}
    }
      	
		
	/**
	 * 
	 */
	public void write(boolean wantHermitian)
        {
    	
    		int numMultiControls = _numBits - _numGits -1;	
    		int numAngles = _radAngles.length();
    		
    		TFControls  centerControls = new TFControls(_numBits, 1);
        	TFControls  sideControls = 
        		new TFControls(_numBits, numMultiControls);
        	boolean[] boolAngles = new boolean [numAngles];
	        int k, b;
	        long x;
	        double fraction, ang;
	        int bitPos = numMultiControls;
        	for(k=1; k<=_bitPrecision; k++){
        		bitPos++;
        		centerControls.setControl(0, bitPos, true);      		
        		for(b=0; b<numAngles; b++){
        			fraction = Globals.between0and1(_radAngles.get(b)
        							/(2*Math.PI));
        			x = (long)(fraction*(1<<k));
        			boolAngles[b] = ((x%2)==1);
        		}
        		
        		writeOmega(bitPos, boolAngles, sideControls);
        		ang = 2*Math.PI/(1<<k);
        		writeControlledOneBitGate(
        		   numMultiControls,//targetBitPosition
        		   centerControls,
        		   new OneBitRot(wantHermitian?-ang:ang, OneBitRot.Y_AXIS));
        		writeOmega(bitPos, boolAngles, sideControls);
        		
        	}
        	
        }
	/**
	 * 
	 */
	public void write(){
		write(false);
	}
	/**
	 * 
	 */
	public void writeHermitian(){
		write(true);
    }

	/**
	 * 
	 */	
	private void writeOmega(int bitPos, 
			boolean[] boolAngles, TFControls sideControls){
		int numAngles = boolAngles.length;
		int numMultiControls = sideControls.getNumControls();
		int b,c;
		for(b=0; b<numAngles; b++){
			if(boolAngles[b]){
				for(c=0; c<numMultiControls; c++){
					sideControls.setControl(c,numMultiControls-c-1,
							((b>>(numMultiControls-c-1))&1)==1);
				}
        		writeControlledOneBitGate(
             		   bitPos,//target
             		   sideControls,
             		   new SigX());
			}
		}
	}
	/**
	 * 
	 */ 
	public static void main(String[] args){
		String	filePrefix = "orac";
		
 
	   int numMultiControls = 1;
	   int numGits = 6;
	   int bitPrecision = 6;
	   boolean noErrorAngles = false;

/*	   
 	   int numMultiControls = 3;
	   int numGits = 4;
	   int bitPrecision = 3;
	   boolean noErrorAngles = true;
*/	  

 


	   int numNonGits = numMultiControls + 1;
 	   int numBits = numNonGits + numGits;
 	   int numAngles = 1<<numMultiControls;
 	   double[]	radAngles = new double[numAngles];
 	   double 	twoPi = 2*Math.PI;
 	   int ranBit;
	   for(int j=0; j<numAngles; j++){
		   if(noErrorAngles){
			   radAngles[j] = 0;
			   for(int p=1; p<=bitPrecision; p++){
				   ranBit = (int)(2*Math.random());
				   radAngles[j] += (twoPi*ranBit)/(1<<p);
			   }
		   }else{
			   radAngles[j] = (Math.random() -.5)*23;
		   }
		   System.out.println("j, radAngles = " + j +", "+ radAngles[j]);
	   }

       new MultiplexorOracularSEO_writer(
           filePrefix, 
           numBits,
           new DoubleArray(radAngles),
           numGits,
	  	   bitPrecision);
       
 	  	MultiplexorUinMat   Uin = new MultiplexorUinMat(numNonGits, 
	  			new DoubleArray(radAngles));
	  	System.out.println("exact:");
	  	System.out.println(Uin);
	  	
    	SEO_readerMu	reader = new SEO_readerMu(
    	    filePrefix,
    	    numBits, 
    	    false); // useSparse
    	int lenOfSEO = reader.getLenOfSEO();
    	
    	System.out.println("approx:");
    	//System.out.println(reader);
        int r, c, max = 1<<numNonGits;
        String output="";
        for(r=0; r<max; r++){
            for(c=0; c<max; c++){
                output += reader.getEntry(r, c).re + " ";
            }
            output += '\n';
        }
        System.out.print(output);
   	
    	double error = 0.0;
    	int row, col;
    	int numRows = 1<<numNonGits;
    	for(row=0; row<numRows; row++){
    		for(col=0; col<numRows; col++){
    			error += Complex.abs(Complex.minus(
    					Uin.getEntry(row,col), 
    					reader.getEntry(row,col)
    					));
    		}
    	}
    	
    	System.out.println("Number of Bits. = " + numBits);
        System.out.println("Number of Elem. Ops. = " + lenOfSEO);
        System.out.printf("Error = %e\n", error);
   	
	}
}
