import java.util.*;
import java.io.*;

/**
 * CB_Net = Classical Bayesian Network
 * This class reads files called
 * states.txt, parents.txt and probs.txt
 * which together fully specify a classical Bayesian network.
 * The class can also calculate various 
 * quantities associated with the Bayesian net,
 * and write various files about the Bayesian net.
 */


public class CB_Net{
	private Node[] 			_nodes;
	private NamesList 		_nodeNamesList;
//	private int 			_numNits;//nit = node bit
	private Nit[] 			_nits;
	/**
	 * 
	 */		
	public CB_Net(
			String 	statesFileName,
			String 	parentsFileName)
	{//default uniform Prob matrix	
		readStatesFile(statesFileName);
		readParentsFile(parentsFileName);
		int numNodes = _nodes.length;
		int numRows, numCols;
		for(int j=0; j<numNodes; j++){
			numRows = _nodes[j].getNumStates();
			numCols = _nodes[j].getProbMatNumCols();
			_nodes[j].setProbMat(numRows, numCols);		
		}		
	}
	/**
	 * 
	 */	
	public CB_Net(
		String 	statesFileName,
		String 	parentsFileName, 
		String 	probsFileName)
	{		
		readStatesFile(statesFileName);
		readParentsFile(parentsFileName);
		readProbsFile(probsFileName);
	}
	/**
	 * 
	 */	
	private void setNodeMBlankets(){
		//markov blanket of focus node consists of:
		//(1)parents of focus node
		//(2)children of focus node
		//(3)other(non-focus node) parents of children (mates) of focus node
		int numNodes = _nodes.length;
		int numParents;
		int j, k;
		int nd, pa;
		Object[] blankets = new Object[numNodes];
		for(nd=0; nd<numNodes; nd++){
			blankets[nd] = new TreeSet<Integer>();
			numParents = _nodes[nd].getNumParents();
			for(j=0; j<numParents; j++){//adding parents
				pa = _nodes[nd].getParent(j);
				((TreeSet<Integer>)blankets[nd]).add(pa);
			}	
		}
		
		int ch, ma;
		for(ch=0; ch<numNodes; ch++){
			numParents = _nodes[ch].getNumParents();
			for(j=0; j<numParents; j++){//adding children
				nd = _nodes[ch].getParent(j);
				((TreeSet<Integer>)blankets[nd]).add(ch);
				for(k=0; k<numParents; k++){
					//adding mates (i.e., other parents of children)
					ma = _nodes[ch].getParent(k);
					if(ma!=nd){
						((TreeSet<Integer>)blankets[nd]).add(ma);
					}
				}
			}		
		}
		for(nd=0; nd<numNodes; nd++){
			Integer[] mb = ((TreeSet<Integer>)blankets[nd]).
				toArray(new Integer[0]);
			int len = mb.length;
			int[] blanket = new int[len];
			for( j=0; j<len; j++){			
				blanket[j] = mb[j];
			}
			_nodes[nd].setMBlanket(blanket);
		}
	}
	/**
	 * 
	 */		
	private void setNits(){
		int nd, numNodes = _nodes.length;
		int numNodeStates;
		int numOwnedNits, totalNits;
		
		totalNits = 0;
		for(nd=0; nd<numNodes; nd++){
			numNodeStates = _nodes[nd].getNumStates();
			numOwnedNits=0;
			_nodes[nd].setFirstOwnedNit(totalNits);
			while((1<<numOwnedNits)<numNodeStates){
				numOwnedNits++;
				totalNits++;
			}
			_nodes[nd].setNumOwnedNits(numOwnedNits);
		}
		int numNits = totalNits;
		
		_nits = new Nit[numNits];
		int ni, t;
		ni=0;
		for(nd=0; nd<numNodes; nd++){
			numOwnedNits = _nodes[nd].getNumOwnedNits();
			for(t=0; t<numOwnedNits; t++){
				_nits[ni] = new Nit(nd);
				ni++;
			}
		}
		
		for(ni=0; ni<numNits; ni++){
			setNitMBlanket(ni);
		}
		
	}	
	/**
	 * 
	 */		
	public void setNitMBlanket(int focusNit)	
	{		
		int 	ownerNode = _nits[focusNit].getOwnerNode();
		int 	blaNd, numBlanketNodes; 
		int 	blaNit, numBlanketNits;
		Node 	blanketNode;
		int[] 	nitBlanket;
		int 	numOwnedNits, firstOwnedNit;
		int 	t;

		numBlanketNodes =_nodes[ownerNode].getMBlanketSize();
		numBlanketNits=0;
		for(blaNd=0; blaNd<numBlanketNodes; blaNd++){
			blanketNode = _nodes[_nodes[ownerNode].getMBlanketNode(blaNd)];
			numBlanketNits += blanketNode.getNumOwnedNits();
		}
		//include in nit markov blanket, the other nits owned by owner node
		numBlanketNits += _nodes[ownerNode].getNumOwnedNits() - 1;
		nitBlanket = new int[numBlanketNits];
		
		blaNit = 0;
		numOwnedNits = _nodes[ownerNode].getNumOwnedNits();
		firstOwnedNit = _nodes[ownerNode].getFirstOwnedNit();
		for(t=0; t<numOwnedNits; t++){
			if(firstOwnedNit + t != focusNit){
				nitBlanket[blaNit] = firstOwnedNit + t;
				blaNit++;
			}
		}
		blaNit = numOwnedNits - 1;
		for(blaNd=0; blaNd<numBlanketNodes; blaNd++){
			blanketNode = _nodes[_nodes[ownerNode].getMBlanketNode(blaNd)];
			numOwnedNits = blanketNode.getNumOwnedNits();
			firstOwnedNit = blanketNode.getFirstOwnedNit();
			for(t=0; t<numOwnedNits; t++){
				nitBlanket[blaNit] = firstOwnedNit + t;
				blaNit++;
			}
		}
		Arrays.sort(nitBlanket);
		_nits[focusNit].setNitMBlanket(nitBlanket);
			
	}
	/**
	 * 
	 */	
	private void readStatesFile(String statesFileName)
	{
		String 	nodeName;
		ArrayList<String> nodeNames = new ArrayList<String>(100);
		ArrayList<String> stateNames = new ArrayList<String>(10);
		ArrayList<Node> nodes = new ArrayList<Node>(100);
		try{
			Scanner input = new Scanner(new File(statesFileName));
			while(input.hasNext("#")){
				input.next("#");//skip hash
				nodeName = input.next();
				nodeNames.add(nodeName);
				while(input.hasNext()&&!input.hasNext("#")){
					stateNames.add(input.next());					
				}
				nodes.add(
					new Node(nodeName, stateNames.toArray(new String[0])));
				stateNames.clear();			
			}
			input.close();
		}
		catch(FileNotFoundException ex){
			System.out.println("\n" + ex.toString());
		}
		
		int numNodes = nodes.size();
		_nodes = new Node[numNodes];
		for(int j=0; j<numNodes; j++){
			_nodes[j] = nodes.get(j);
		}
		_nodeNamesList = new NamesList(nodeNames.toArray(new String[0]));
	}
	/**
	 * 
	 */	
	private void readParentsFile(String parentsFileName)
	{	
		
		String	nodeName;
		ArrayList<Integer> parentsAL = new ArrayList<Integer>(20);
		int j, numParents, nodeFillNum = -1;
		int[] parents, parentDegens;
		try{
			Scanner input = new Scanner(new File(parentsFileName));
			while(input.hasNext("#")){
				nodeFillNum++;
				input.next("#");//skip hash
				nodeName = input.next();
				Throw.If(!nodeName.equals(_nodes[nodeFillNum].getName()),
				 	"While reading Parents File, " +
				 	"encountered the illegal focus node name:\n" +
				 	"\"" + nodeName + "\"\n" +
				 	"Note that focus node names " +
				 	"(i.e, strings immediately after a hash) " +
				 	"must be in the same order " +
				 	"in Parents File, States File, and Probs File.");
				while(input.hasNext()&&!input.hasNext("#")){
					parentsAL.add(_nodeNamesList.getFillNum(input.next()));
				}
				numParents = parentsAL.size();
				parents = new int[numParents];
				parentDegens = new int[numParents];
				for(j=0;j<numParents;j++){
					parents[j] = parentsAL.get(j);
					parentDegens[j] = _nodes[parents[j]].getNumStates();
				}				
				_nodes[nodeFillNum].setParents(parents);
				_nodes[nodeFillNum].setParentDegens(parentDegens);
				parentsAL.clear();
			}
			input.close();
		}
		catch(FileNotFoundException ex){
			System.out.println("\n" + ex.toString());
		}
		setNodeMBlankets();
		setNits();
	}
	/**
	 * 
	 */
	private void readProbsFile(String probsFileName){
			
		String	nodeName;
		int	row, numStates;
		int col, numParents;
		int[] 	colArray;
		double 	prob;
		int nodeFillNum = -1;
		try{
			Scanner input = new Scanner(new File(probsFileName));
			while(input.hasNext("#")){
				nodeFillNum++;
				input.next("#");//skip hash
				nodeName = input.next();
				Throw.If(!nodeName.equals(_nodes[nodeFillNum].getName()),
				 	"While reading Probs File, " +
				 	"encountered the illegal focus node name:\n" +
				 	"\"" + nodeName + "\"\n" +
				 	"Note that focus node names " +
				 	"(i.e, strings immediately after a hash) " +
				 	"must be in the same order " +
				 	"in Parents File, States File, and Probs File.");
				numStates = _nodes[nodeFillNum].getNumStates();
				Object[] probMat = new Object[numStates];
				for(int j=0; j<numStates; j++){
					probMat[j] = new ArrayList<IntDoublePair>(50);
				}
				numParents = _nodes[nodeFillNum].getNumParents();	
				colArray = new int[numParents];	
				while(input.hasNext()&&!input.hasNext("#")){
					row = _nodes[nodeFillNum].getStateNamesList().
						getFillNum(input.next());		
					for(int j=0; j<numParents; j++){
						colArray[j] = _nodes[_nodes[nodeFillNum].getParent(j)]
						   .getStateNamesList().getFillNum(input.next());
					}
					col = _nodes[nodeFillNum].getProbMatColNum(colArray);
					prob = Double.parseDouble(input.next()); 
					((ArrayList<IntDoublePair>)probMat[row]).add(
							new IntDoublePair(col, prob));
				}
				_nodes[nodeFillNum].setProbMat(probMat);
				for(int j=0; j<probMat.length; j++){
					((ArrayList<IntDoublePair>)probMat[j]).clear();
				}
			}
			input.close();
		}
		catch(FileNotFoundException ex){
			System.out.println("\n" + ex.toString());
		}
	}
	/**
	 * 
	 */
	public Node getNode(String nodeName){	
		return _nodes[_nodeNamesList.getFillNum(nodeName)];	
	}
	/**
	 * 
	 */
	public Node getNode(int nodeFillNum){	
		return _nodes[nodeFillNum];	
	}
	/**
	 * 
	 */
	public int getNumNodes(){
		return _nodes.length;
	}
	/**
	 * 
	 */
	public double getProb(int[] nodeStates)
	{	
		double prob;
		int nd, numNodes = _nodes.length;		
		int pa, numParents;
		int[] colArray;
		int col, row;
		
		prob = 1;
		for (nd=0; nd<numNodes; nd++){
			numParents = _nodes[nd].getNumParents();
			colArray = new int[numParents];
			for(pa=0; pa<numParents; pa++){
				colArray[pa] = nodeStates[_nodes[nd].getParent(pa)];
			}
			row = nodeStates[nd];
			col = _nodes[nd].getProbMatColNum(colArray);
			prob *= _nodes[nd].getProb(row, col);
		}
		return prob;
	}
	/**
	 * 
	 */
	public double getProb(boolean[] nitStates)
	{
		int[] nodeStates;
		try{
			nodeStates = getNodeStatesFromNitStates(nitStates);
		}
		catch(Exception_NitStateWithoutNodeState ex){
			return 0;
		}
		
		return getProb(nodeStates);
		
	}
	/**
	 * 
	 */
	public int[] getNodeStatesFromNitStates(boolean[] nitStates)
		throws Exception_NitStateWithoutNodeState
	{
		int nd, numNodes = _nodes.length;
		int numOwnedNits, firstOwnedNit;
		int t;
		int sum, coef;
		int[] nodeStates = new int[numNodes];
		
		for(nd=0; nd<numNodes; nd++){
			numOwnedNits = _nodes[nd].getNumOwnedNits();
			firstOwnedNit = _nodes[nd].getFirstOwnedNit();
			sum = 0;
			coef = 1;
			for(t=0; t<numOwnedNits; t++){
				sum += coef*(nitStates[firstOwnedNit + t]==false?0:1);
				coef *= 2;
			}
			if(sum>=_nodes[nd].getNumStates()){
				throw new Exception_NitStateWithoutNodeState();
			}
			nodeStates[nd] = sum;
		}
		return nodeStates;
		
	}
	/**
	 * 
	 */
	public boolean[] getNitStatesFromNodeStates(int[] nodeStates)
	{
		int nd, numNodes = _nodes.length;
		int numNits = _nits.length;
		int numOwnedNits, firstOwnedNit;
		int t;
		boolean[] nitStates = new boolean[numNits];
		
		int nodeState;
		for(nd=0; nd<numNodes; nd++){
			numOwnedNits = _nodes[nd].getNumOwnedNits();
			firstOwnedNit = _nodes[nd].getFirstOwnedNit();
			for(t=0; t<numOwnedNits; t++){
				nodeState = nodeStates[nd];
				nitStates[firstOwnedNit + t]= 
					((nodeState>>t)&1)==1?true:false;
			}
		}
		return nitStates;
		
	}
	/**
	 * 
	 */
	public double getProbOfOneNitGivenOthers(
		int focusNit, 
		boolean[] nitStates)
	{
		int ni, numNits = _nits.length;
		boolean[]  tempoNitStates = new boolean[numNits];
		double prob0, prob1, denom;

		for(ni=0; ni<numNits; ni++){
			tempoNitStates[ni] = nitStates[ni];
		}
		tempoNitStates[focusNit] = false;
		prob0 = getProb(tempoNitStates);
		tempoNitStates[focusNit] = true;
		prob1 = getProb(tempoNitStates);
		denom = prob0 + prob1;
		double ans;
		if(denom<1e-9){
			ans = (nitStates[focusNit]==false?1:0);
		}else{
			ans = (nitStates[focusNit]==false?prob0:prob1)/denom;
		}
		return ans;
	}
	/**
	 * 
	 */
	public double getM1_entry(
		boolean[] yNitStates,
		boolean[] xNitStates)
	{//entries of markov chain M_1(y|x)
		int ni, numNits = _nits.length;
		boolean[] tempoNitStates = new boolean[numNits];		
		for(ni=0; ni<numNits; ni++){
			tempoNitStates[ni] = xNitStates[ni];		
		}
		
		double prob = 1;
		for(ni=0; ni<numNits; ni++){
			tempoNitStates[ni] = yNitStates[ni];
			prob *= getProbOfOneNitGivenOthers(ni, tempoNitStates);
		}
		return prob;
	}
	/**
	 * 
	 */
	public double getM2_entry(
		boolean[] yNitStates,
		boolean[] xNitStates)
	{//entries of markov chain M_2(y|x)
		int ni, numNits = _nits.length;
		boolean[] tempoNitStates = new boolean[numNits];		
		for(ni=0; ni<numNits; ni++){
			tempoNitStates[ni] = yNitStates[ni];		
		}
		
		double prob = 1;
		for(ni=0; ni<numNits; ni++){
			prob *= getProbOfOneNitGivenOthers(ni, tempoNitStates);
			tempoNitStates[ni] = xNitStates[ni];
		}
		return prob;
	}
	
/*	
	//The following code can be used if you do not
	//want to use the Nit class.

	public int getNumNits(){
		return _numNits;
	}
	
	public int getNitOwnerNode(int nitFillNumber)
	{
		int nd=-1, numNodes = _nodes.length;
		int firstOwnedNit;
		while(nd<numNodes-1){
			nd++;
			firstOwnedNit = _nodes[nd].getFirstOwnedNit();
			if(firstOwnedNit > nitFillNumber){
				nd--;
				break;
			}
		}
		return nd;	
	}
	
	public int[] getNitMBlanket(int focusNit)	
	{		
		int 	ownerNode = getNitOwnerNode(focusNit);
		int 	blaNd, numBlanketNodes; 
		int 	blaNit, numBlanketNits;
		Node 	blanketNode;
		int[] 	nitBlanket;
		int 	numOwnedNits, firstOwnedNit;
		int 	t;

		numBlanketNodes =_nodes[ownerNode].getMBlanketSize();
		numBlanketNits=0;
		for(blaNd=0; blaNd<numBlanketNodes; blaNd++){
			blanketNode = _nodes[_nodes[ownerNode].getMBlanketNode(blaNd)];
			numBlanketNits += blanketNode.getNumOwnedNits();
		}
		//include in nit markov blanket, the other nits owned by owner node
		numBlanketNits += _nodes[ownerNode].getNumOwnedNits() - 1;
		nitBlanket = new int[numBlanketNits];
		
		blaNit = 0;
		numOwnedNits = _nodes[ownerNode].getNumOwnedNits();
		firstOwnedNit = _nodes[ownerNode].getFirstOwnedNit();
		for(t=0; t<numOwnedNits; t++){
			if(firstOwnedNit + t != focusNit){
				nitBlanket[blaNit] = firstOwnedNit + t;
				blaNit++;
			}
		}
		blaNit = numOwnedNits - 1;
		for(blaNd=0; blaNd<numBlanketNodes; blaNd++){
			blanketNode = _nodes[_nodes[ownerNode].getMBlanketNode(blaNd)];
			numOwnedNits = blanketNode.getNumOwnedNits();
			firstOwnedNit = blanketNode.getFirstOwnedNit();
			for(t=0; t<numOwnedNits; t++){
				nitBlanket[blaNit] = firstOwnedNit + t;
				blaNit++;
			}
		}
		Arrays.sort(nitBlanket);
		return nitBlanket;
			
	}
*/
	/**
	 * 
	 */	
	public int getNumNits(){
		return _nits.length;
	}

	public  Nit[]  getNits(){
		return _nits;
	}
	/**
	 * 
	 */
	public void writeStatesFile(String statesFileName){
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(statesFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
 	
		int numNodes = _nodes.length;
		int numStates;
		int nd, st;
		for(nd=0; nd<numNodes; nd++){
			output.println("#\t" + _nodes[nd].getName());
			numStates = _nodes[nd].getNumStates();
			for(st=0; st<numStates; st++){
				output.println(_nodes[nd].getStateNamesList().getName(st));
			}			
		}		
		output.close();		
	}
	/**
	 * 
	 */
	public void writeWinBUGSFiles(
			String relPath,
			String netName){
		PrintWriter outBUGS = null;
	   	try{
	   		outBUGS = new PrintWriter(new File(relPath + 
	   				netName + "_BUGS.txt"));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
    	
    	//IMP: R, BUGS variable names cannot start with integer!!
    	//write model specification 
    	
       	outBUGS.println("model {");
       	
		int numNodes = _nodes.length;
		String nodeName = "";
		int numParents;
		int nd, pa;	
		String str = "";
		int coef;
		Node paNode;
		for(nd=0; nd<numNodes; nd++){
			nodeName = _nodes[nd].getName();
			outBUGS.print(
				"\t" + nodeName + " ~ " + "dcat(p_" + nodeName + "[");
			numParents = _nodes[nd].getNumParents();
			if(numParents!=0){
				outBUGS.print("input_" + nodeName + ", ");
			}
			outBUGS.println("])");
			/*quote from WinBUGS manual(version 14, page 16):
			a ~ dcat(p.a[1:2])
			b ~ dcat(p.b[1:2])
			c ~ dcat(p.c[1:2])
			d ~ dcat(p.d[a, b, c, 1:2]
			WinBUGS will not permit this, 
			and so the index must be explicitly calculated:
			d ~ dcat(p[k, 1:2])
			k <- 8 * (a - 1) + 4 * (b - 1) + c
			*/
			
			if(numParents!=0){	
				pa = 0;
				paNode = _nodes[_nodes[nd].getParent(pa)];
				str = paNode.getName();
				coef = paNode.getNumStates();
				pa++;
				while(pa<numParents){
					paNode = _nodes[_nodes[nd].getParent(pa)];
					str = coef + "*(" +
						paNode.getName() + 
						" - 1) + " + str;
					coef *= paNode.getNumStates();
					pa++;
				}
				outBUGS.println("\t\tinput_" + nodeName + " <- " + str);
			}			
		}
		outBUGS.println("}");
		outBUGS.close();

		PrintWriter outR = null;
	   	try{
	   		outR = new PrintWriter(new File(relPath,
	   				netName + "_R1.txt"));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
    	//load necessary library
    	outR.print("library(\"R2WinBUGS\")\n\n");
    	
    	outR.println("#set correct working directory");
    	outR.print(
    	"setwd(\"C:\\\\WINDOWS\\\\Desktop\\\\Backed_up\\\\R-results\")\n\n"
    	);
    			
    	

    	//define distributions of categorical nodes
		int numStates, numCols;
		int st, col;
		
		outR.print("#WinBUGS requires varying out-states first, but" +
		"\n#early versions of R package \"R2WinBUGS\" " +
		"have a bug that requires\n"+
		"#varying in-states first\n\n");

		boolean varyOutStatesFirst;
		for(int ordering=0; ordering<2; ordering++){
		varyOutStatesFirst=(ordering==0?false:true);
		outR.println("#define distributions of categorical nodes");
		if(varyOutStatesFirst){
			outR.println("#varying out-states first (WinBUGS convention)");
		}else{
			outR.println("#varying in-states first");
		}
		outR.println("data = list(");
		for(nd=0; nd<numNodes; nd++){
			numParents = _nodes[nd].getNumParents();
			nodeName = _nodes[nd].getName();
			if(numParents==0){
				outR.println("\tp_" + nodeName + " = c(");
			}else{
				outR.println("\tp_" + nodeName + 
						" = structure(.Data = c(");
			}
			numStates = _nodes[nd].getNumStates();
			numCols =  _nodes[nd].getProbMatNumCols();
			if(varyOutStatesFirst){
				for(col=0; col<numCols; col++){
					outR.print("\t\t");
					for(st=0; st<numStates; st++){
						outR.print(_nodes[nd].getProb(st, col));
						outR.print((col==numCols-1&&st==numStates-1)?"":",\t");
					}
					outR.print("\n");
				}
			}else{
				for(st=0; st<numStates; st++){
					outR.print("\t\t");
					for(col=0; col<numCols; col++){
						outR.print(_nodes[nd].getProb(st, col));
						outR.print((col==numCols-1&&st==numStates-1)?"":",\t");
					}
					outR.print("\n");
				}
			}
			if(numParents==0){
				outR.print("\t)");
			}else{
				outR.print("\t), .Dim = c("); 
				outR.print(numCols + "," + numStates);
				outR.print("))");
			}
			outR.print(nd==numNodes-1?"\n":",\n");
		}//nd loop
		outR.print(")\n\n");
		}//ordering 
		
		//define parameters and their inits
		//(assuming no observed nodes, so all nodes are parameters)
		outR.println("parameters = list(");
		for(nd=0; nd<numNodes; nd++){
			outR.print("\t\"" + _nodes[nd].getName() + "\"");
			outR.print(nd==numNodes-1?"":",\n");
		}
		outR.print("\n)\n\n");
		
		outR.println("#fixed inits");
		outR.print("inits = list(\n");
		for(int j=0;j<3;j++){
			outR.print("\tlist(");
			for(nd=0; nd<numNodes; nd++){
				outR.print(_nodes[nd].getName() + "=1");
				outR.print(nd==numNodes-1?"":",");
			}
			outR.print(j==2?")\n":"),\n");
		}
		outR.print(")\n\n");
		
		outR.println("#random inits");
		outR.println("#runif = random uniform (num samples, min, max)");
		outR.println("inits = function() list(");
		for(nd=0; nd<numNodes; nd++){
			outR.print("\t" + _nodes[nd].getName() +
					" = 1 + trunc(runif(1, 0, " +
					_nodes[nd].getNumStates() + "))");
			outR.print(nd==numNodes-1?"":",\n");
		}
		outR.print("\n)\n\n");
		
		//bugs() calls
		outR.println("binSize = 50");
		outR.println("numBins = 50");
		outR.println("numSam = binSize*numBins");
		outR.println("bugOut = bugs(\n" +
			"\tdata, inits, parameters,\n" +
			"\t\"" + netName + "_BUGS.txt\",\n" +
			"\tn.chains=3,\n\tn.iter=numSam,\n" +
			"\tn.thin=1,\n\tn.burnin=0,\n" +
			"\tcodaPkg=TRUE,\n\t#codaPkg=FALSE,\n" +
			"\tdebug=TRUE,\n" +
		    "\tDIC=FALSE)  #no nodes are observed so DIC is undefined"
		);
		
		outR.close();

	}
	/**
	 * 
	 */
	public  void writeParentsFile(String parentsFileName){
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(parentsFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
 	
		int numNodes = _nodes.length;
		int numParents;
		int nd, pa;
		for(nd=0; nd<numNodes; nd++){
			output.println("#\t" + _nodes[nd].getName());
			numParents = _nodes[nd].getNumParents();
			for(pa=0; pa<numParents; pa++){
				output.println(_nodes[_nodes[nd].getParent(pa)].getName());
			}	
		}		
		output.close();
	}		
	/**
	 * 
	 */
	public  void writeProbsFileVaryingOutStatesFirst(String probsFileName){
		//probMatrix(outState|inState)
		
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(probsFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
 	
		int numNodes = _nodes.length;
		int numStates, numCols, numParents;
		int nd, st, col, pa;
		int[] colArray;
		double prob;
		for(nd=0; nd<numNodes; nd++){
			output.println("#\t" + _nodes[nd].getName());
			numStates = _nodes[nd].getNumStates();
			numCols =  _nodes[nd].getProbMatNumCols();
			numParents = _nodes[nd].getNumParents();
			for(col=0; col<numCols; col++){
				colArray = _nodes[nd].getProbMatColArray(col);
				for(st=0; st<numStates; st++){
					prob = _nodes[nd].getProb(st, col);
					if(prob>1e-9){
						output.print(
							_nodes[nd].getStateNamesList().getName(st));
						for(pa=0; pa<numParents; pa++){
							output.print("\t" +
								_nodes[_nodes[nd].getParent(pa)].
								getStateNamesList().getName(colArray[pa]));
						}
						output.println("\t" + prob);
					}
				}						
			}
		}		
		output.close();
	}
	/**
	 * 
	 */
	public  void writeProbsFileVaryingInStatesFirst(String probsFileName){
		//probMatrix(outState|inState)
		
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(probsFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
 	
		int numNodes = _nodes.length;
		int numStates, lenRaggedRow, numParents;
		int nd, st, rcol, pa;
		int[] colArray;
		int[] intsOfRow;
		double[] doublesOfRow;
		double prob;
		for(nd=0; nd<numNodes; nd++){
			output.println("#\t" + _nodes[nd].getName());
			numStates = _nodes[nd].getNumStates();
			numParents = _nodes[nd].getNumParents();
			for(st=0; st<numStates; st++){
				intsOfRow = _nodes[nd].getProbMatIntsOfRow(st);
				doublesOfRow = _nodes[nd].getProbMatDoublesOfRow(st);
				lenRaggedRow = intsOfRow.length;
				for(rcol=0; rcol<lenRaggedRow; rcol++){
					output.print(_nodes[nd].getStateNamesList().getName(st));
					colArray = _nodes[nd].getProbMatColArray(intsOfRow[rcol]);			
					prob = doublesOfRow[rcol];
					for(pa=0; pa<numParents; pa++){
						output.print("\t" +
							_nodes[_nodes[nd].getParent(pa)].
							getStateNamesList().getName(colArray[pa]));
					}
					output.println("\t" + prob);
				}						
			}
		}		
		output.close();
	}
	/**
	 * 
	 */
	public  void writeProbsFile(
			String 	probsFileName,		//output file
			boolean varyInStatesFirst) //true here is faster
	{		
		if(varyInStatesFirst){
			writeProbsFileVaryingInStatesFirst(probsFileName);
		}else{
			writeProbsFileVaryingOutStatesFirst(probsFileName);
		}
	}
	/**
	 * 
	 */
	public  void writeMBlanketsFile(String blanketsFileName){
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(blanketsFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }
 	
		int j, numNodes = _nodes.length;
		int k, blanketSize;
		for(j=0; j<numNodes; j++){
			output.println("#\t" + _nodes[j].getName());
			blanketSize = _nodes[j].getMBlanketSize();
			for(k=0; k<blanketSize; k++){
				output.println(_nodes[_nodes[j].getMBlanketNode(k)].getName());
			}		
		}
		output.close();
	}
	/**
	 * 
	 */
	public  void writeNitsFile(String nitsFileName){
		PrintWriter output = null;
	   	try{
	   		output = new PrintWriter(new File(nitsFileName));
    	}
    	catch(FileNotFoundException ex){
    		System.out.println("\n" + ex.toString());
        }	
    	int ni, numNits = _nits.length;
    	int bla, nitBlanketSize;
    	int[] nitBlanket;
    	for(ni=0; ni<numNits; ni++){
    		output.println("#\t" + ni);
    		output.println("ownerNode: " + 
    			getNode(_nits[ni].getOwnerNode()).getName());
    		output.print("nit blanket: ");
    		nitBlanket = _nits[ni].getNitMBlanket();
    		nitBlanketSize = nitBlanket.length;
    		for(bla=0; bla<nitBlanketSize; bla++){
    			output.print("(" + nitBlanket[bla] + " in " +
    				getNode(_nits[nitBlanket[bla]].getOwnerNode()).getName() 
    				+ ")");	
    		}
    		output.print("\n");
    	}

		output.close();
	}
	/**
	 * 
	 */
	public String[] nodeStatesArrayToStrings(int[] nodeStates){
		int nd, numNodes = _nodes.length;
		int	nodeState;
		int t, numOwnedNits, firstOwnedNit;	
		String[] strArray = new String[numNodes];
		String str="";
		boolean[] nitStates = getNitStatesFromNodeStates(nodeStates);
		for(nd=0; nd<numNodes; nd++ ){
			nodeState = nodeStates[nd];
			numOwnedNits = _nodes[nd].getNumOwnedNits();
			firstOwnedNit = _nodes[nd].getFirstOwnedNit();
			str = _nodes[nd].getName() + ":\t" + 
				_nodes[nd].getStateNamesList().getName(nodeState) +
				"(";
			for(t=numOwnedNits-1; t>=0; t--){
				str +=(nitStates[firstOwnedNit + t]==false?0:1);			
			}
			str += ")" + nodeState;	
			strArray[nd] = new String(str);
		}
		return strArray;	
	}
	/**
	 * 
	 */
	public void checkNodeStatesArray(int[] nodeStates){
		int j, numNodes = _nodes.length;
		int	numStates;
		Throw.If(nodeStates.length!=numNodes,
			"bad nodeStates: wrong number of nodes");
		for(j=0; j<numNodes; j++ ){
			numStates = _nodes[j].getNumStates();
			Throw.If(nodeStates[j]<0 || nodeStates[j]>=numStates,
				"bad nodeStates: out of range state for a node");
		}	
	}
	/**
	 * 
	 */
	public int[] getRandomNodeStatesArray(){
		int j, numNodes = _nodes.length;
		int[] nodeStates = new int[numNodes];
		int numStates;
		for(j=0; j<numNodes; j++ ){
			numStates = _nodes[j].getNumStates();
			nodeStates[j] =(int)( Math.random()*numStates);
		}	
		return nodeStates;
	}
	/**
	 * 
	 */
	public class Exception_NitStateWithoutNodeState extends Exception{}
	/**
	 * 
	 */
	public static void main(String[] args){
		
		//String 	netName = "2nodes";
		//String 	netName = "3nodes";
		//String 	netName = "4nodeFullyConnected";
		//String 	netName = "Asia";
		//String 	netName = "Tree4Level";
		String 	netName = "net9-2";

		String 	relPath = "../Quibbs/" + netName + "/";
		String 	statesFileName = "states.txt";
		String 	parentsFileName = "parents.txt";
		
		CB_Net  net;
		int test = 3;

		switch(test){
			case 1:
				net = new CB_Net(relPath + statesFileName, 
						relPath + parentsFileName);
				net.writeProbsFile(netName + "/" + "probsT.txt", 
					true);			//varyInStatesFirst		
		
				net.writeProbsFile(netName + "/" + "probsF.txt", 
					false);			//varyInStatesFirst
				net.writeMBlanketsFile(netName + "/" + "blankets.txt");
				net.writeNitsFile(netName + "/" + "nits.txt");
				break;
			case 2:
				net = new CB_Net(relPath + statesFileName, 
						relPath + parentsFileName,
						relPath + "probs.txt");
				net.writeStatesFile(netName + "/" + "states1.txt"); 
				net.writeParentsFile(netName + "/" + "parents1.txt"); 
				net.writeProbsFile(netName + "/" + "probs1T.txt",
									 true);//varyInStatesFirst);
				net.writeProbsFile(netName + "/" + "probs1F.txt",
									 false);//varyInStatesFirst
				break;
			case 3:
				net = new CB_Net(relPath + statesFileName, 
						relPath + parentsFileName,
						relPath + "probs.txt");
				net.writeWinBUGSFiles("WinBUGSFiles/", netName);
		}
		
	}

}