/**
 * CB_NetQEmb = classical Bayesian net q-Embedding
 * This class outputs the multiplexor angles
 * used to construct a quantum embedding.
 * Its main input is a Bayesian net.
 */
public class CB_NetQEmb extends QEmb{
//	protected	int				_numColBits;
	private DoubleArray[]  	_radAngles1;//in radians, for Markov chain M_1
	private DoubleArray[]  	_radAngles2;//in radians, for Markov chain M_2
	boolean _usingEffiMeth;//using efficient method
	
	/**
	 * 
	 */
	public CB_NetQEmb(CB_Net net, boolean usingEffiMeth){
		
		_usingEffiMeth = usingEffiMeth;
		if(usingEffiMeth){
			_numColBits = net.getNumNits();
			_radAngles1 =  new  DoubleArray [_numColBits];
			_radAngles2 =  new  DoubleArray [_numColBits];
			int j;
			for(j=0; j < _numColBits; j++){//j is numRowBits
				_radAngles1[j] = new EffiDoubleArray(j, net, 1);
				_radAngles2[j] = new EffiDoubleArray(j, net, 2);
			}			
		}else{
			//this choice is more inefficient.
			//It is only used for testing.
			_numColBits = net.getNumNits();
			int numCols  = 1<<_numColBits;
	
			double[][]  probMat1 = new double[numCols][numCols];
			double[][]  probMat2 = new double[numCols][numCols];
			int row, col;
			boolean[] rowArray, colArray;
			for(col=0; col<numCols; col++){
				colArray = Globals.boolArrayFromInt(col, _numColBits);
				for(row=0; row<numCols; row++){
					rowArray = Globals.boolArrayFromInt(row, _numColBits);
					probMat1[row][col] = 
						net.getM1_entry(rowArray, colArray); 
					probMat2[row][col] = 
						net.getM2_entry(rowArray, colArray); 			
				}			
			}
			ProbMatrices.checkColsAddToOne(probMat1);
			ProbMatrices.checkColsAddToOne(probMat2);
			_radAngles1 = buildAngleArraysFromProbMat(probMat1);
			_radAngles2 = buildAngleArraysFromProbMat(probMat2);
		}
		
	}
	/**
	 * 
	 */
	public boolean usingEffiMeth(){
		return _usingEffiMeth;
	}
	/**
	 * 
	 */
	public double getAngle1(int numRowBits, int  flatIndex){
		Throw.If(numRowBits >=  _numColBits, "too many row bits");
		Throw.If(flatIndex >=  1<<(numRowBits +  _numColBits), 
				"angles index is out of bounds");
		return   _radAngles1[numRowBits].get(flatIndex);
	}
	/**
	 * 
	 */
	public double getAngle2(int numRowBits, int  flatIndex){
		Throw.If(numRowBits >=  _numColBits, "too many row bits");
		Throw.If(flatIndex >=  1<<(numRowBits +  _numColBits), 
				"angles index is out of bounds");
		return   _radAngles2[numRowBits].get(flatIndex);
	}	
	/**
	 * 
	 */
	public DoubleArray[] getAngles1(){
		return   _radAngles1;
	}
	/**
	 * 
	 */
	public DoubleArray[] getAngles2(){
  		return   _radAngles2;
	}
	/**
	 * 
	 */
	public boolean[] getUsedAngleBits(
			int numRowBits,
			int mType)
	{
		Throw.If(!_usingEffiMeth, "This method requires that " +
			"_usingEffiMeth==true");
			EffiDoubleArray	effiArray;
		if(mType==1){
			effiArray =  (EffiDoubleArray)_radAngles1[numRowBits];
		}else{//mType==2
			effiArray =  (EffiDoubleArray)_radAngles2[numRowBits];
		}
		return effiArray.getUsedAngleBits();
	}
	/**
	 * 
	 */
	private class EffiDoubleArray extends DoubleArray
	{
		private int 		_numRowBits;
		private CB_Net 		_net;
		private int 		_mType;//either 1 (for M1) or 2 (for M2)
		/**
		 * 
		 */
		public EffiDoubleArray(
				int 	numRowBits, 
				CB_Net 	net, 
				int 	mType)
		{
			Throw.If(mType!=1 && mType!=2, 
				"markovChain must be either 1 or 2");
			_len = 1<<(numRowBits + net.getNumNits());
			_numRowBits = numRowBits;
			_net = net;
			_mType = mType;
			_useNumeric = false; //default
			
		}
		/**
		 * 
		 */
		public double get(int index)//override
		{
			if(_useNumeric==true)return _array[index];
			int numNits = _net.getNumNits();
			int numAngleBits = _numRowBits + numNits;
			boolean[]  mixedArray = 
				Globals.boolArrayFromInt(index, numAngleBits);
			boolean[] yArray = new boolean[numNits];
			boolean[] xArray = new boolean[numNits];
			boolean[] nitStates = new boolean[numNits];
			int focusNit, j;
			if(_mType==1){
				focusNit = _numRowBits;
				for(j=numAngleBits-1; j>=numNits; j--){
					//yArray[0] = mixedArray[numNits]
					yArray[j-numNits] = mixedArray[j]; 
				}
				for(j=numNits-1; j>=0; j--){
					//xArray[0] = mixedArray[0]
					xArray[j] = mixedArray[j];
				}
				for(j=numNits-1; j>=_numRowBits; j--){
					nitStates[j] = xArray[j];			
				}
				for(j=_numRowBits-1; j>=0; j--){
					nitStates[j] = yArray[j];	
				}
			}else{//_mtype==2
				focusNit = numNits - _numRowBits - 1;
				for(j=numAngleBits-1; j>=numNits; j--){
					//yArray[numNits -1] = mixedArray[numNits];
					yArray[2*numNits -1 - j] = mixedArray[j];
				}
				for(j=numNits-1; j>=0; j--){
					//xArray[numNits-1] = mixedArray[0];
					xArray[numNits-1-j] = mixedArray[j];
				}
				for(j=numNits-1; j>=numNits-_numRowBits; j--){
					nitStates[j] = yArray[j];			
				}
				for(j=numNits-_numRowBits-1; j>=0; j--){
					nitStates[j] = xArray[j];	
				}
			}
			
			nitStates[focusNit] = false;
			double prob = _net.getProbOfOneNitGivenOthers(focusNit, nitStates);
			double	x =  Math.sqrt(prob);
			double 	y =  Math.sqrt(1-prob);	
			return -Math.atan2(y,x);		
		}
		/**
		 * 
		 */
		public boolean[] getUsedAngleBits()
		{
			int numNits = _net.getNumNits();
			int numAngleBits = _numRowBits + numNits;
			boolean[] isUsedBit = Globals.T_vec(numAngleBits);
			int j;
			
			//Mark unused x's.			
			//Same for both mTypes.
			for(j=_numRowBits; j>= 0; j--){
				isUsedBit[j] = false;
			}	
			
			
			//Mark as unused those bits that
			//lie outside blanket of focus nit.
			int[] blanket;
			boolean[] nitIsInBlanket;
			int focusNit, blanketLen;
			if(_mType==1){
				focusNit = _numRowBits;
				blanket = _net.getNits()[focusNit].getNitMBlanket();
				nitIsInBlanket = Globals.F_vec(numNits);
				
				blanketLen = blanket.length;
				for(j=0; j<blanketLen; j++){
					nitIsInBlanket[blanket[j]]=true;
				}
				for(j=numAngleBits-1; j>=numNits; j--){
					if(!nitIsInBlanket[j-numNits])isUsedBit[j]=false;
				}
				for(j=numNits-1; j>=0; j--){
					if(!nitIsInBlanket[j])isUsedBit[j]=false;
				}
			}else{//_mType==2
				focusNit = numNits - _numRowBits - 1;
				blanket = _net.getNits()[focusNit].getNitMBlanket();
				nitIsInBlanket = Globals.F_vec(numNits);
				
				blanketLen = blanket.length;
				for(j=0; j<blanketLen; j++){
					nitIsInBlanket[blanket[j]] = true;
				}
				for(j=numAngleBits-1; j>=numNits; j--){
					if(!nitIsInBlanket[2*numNits -1 -j])isUsedBit[j]=false;
				}
				for(j=numNits-1; j>=0; j--){
					if(!nitIsInBlanket[numNits -1 -j])isUsedBit[j]=false;
				}	
			}
					
			return isUsedBit;
		}	
	}
}
