/**
 * This class writes sequence of operations for an
 * Adaptive Fixed-point Grover's Algorithm (AFGA)
 * as explained in my paper arXiv:1001.5200
  */
public abstract class GroverAdaptiveWriter 
{
	//g=gamma
	protected int 		_maxNumSteps;
	protected double	_gam0;//first gamma
	protected double 	_gamTol; //gamma tolerance
	protected double 	_deltaLam;//Lam=lambda

	
	protected double	_gam;//current gamma
	protected double	_alp;//current alpha
	//the goal is to try to reduce abs(_gam) below _gamTol
	//unless we reach _maxNumSteps steps, in which case
	//we stop without attaining our goal.
	protected int _numWriteSteps;
	protected int _numWriteHermitianSteps;

	/**
	 *
	 */	
	public GroverAdaptiveWriter(
		int 	maxNumSteps,
		double	gam0,
		double 	gamTol,
		double 	deltaLam)
	{
		_maxNumSteps = maxNumSteps;
		_gam0 = gam0;
		_gamTol = gamTol;
		_deltaLam = deltaLam;
		Throw.If(gamTol<0, "Gamma tolerance must be positive.");
		
	}
	/**
	 *
	 */	
	public double getGamma(){
		   return _gam;
	}
	/**
	 *
	 */	
	 public int getNumGroverSteps(){
		   return Math.max(_numWriteSteps, _numWriteHermitianSteps);
	 }
	/**
	 *
	 */	
    public void write()
    {	
    	int count;
    	_gam = _gam0;	    	
     	double gamOut = changeAlp();
     	System.out.println("GroverAdaptiveWriter:write()");
     	System.out.println("step, gamma (degs)");
    	for(count=0; 
    		count<_maxNumSteps && Math.abs(_gam)>_gamTol; count++){
	    	System.out.println(count + "\t" + _gam*180/Math.PI);
	    	writeTargetControlledPhase(false);
	    	writeStartControlledPhase(false);
	    	_gam = gamOut;
	    	gamOut = changeAlp();
    	}
    	_numWriteSteps = count;
 	}
    /**
	 *
	 */	
    public void writeHermitian()
    {
    	int count;
    	_gam = _gam0;	    	
     	double gamOut = changeAlp();
     	System.out.println("GroverAdaptiveWriter:writeHermtian()");
     	System.out.println("step, gamma (degs)");
    	for(count=0; 
    		count<_maxNumSteps && Math.abs(_gam)>_gamTol; count++){
	    	System.out.println(count + "\t" +_gam*180/Math.PI);
	    	writeStartControlledPhase(true);
	    	writeTargetControlledPhase(true);
	    	_gam = gamOut;
	    	gamOut = changeAlp();
    	}
    	_numWriteHermitianSteps = count;
    }
    /**
	 * Changes _alp and returns what the next value of _gam should be
	 */	
    private double changeAlp(){
	   
		double cgin = Math.cos(_gam);
		double sgin = Math.sin(_gam);
		double cg0 = Math.cos(_gam0);
		double sg0 = Math.sin(_gam0);
		double cg0minus = Math.cos(_gam0 - _gam);
		double sg0minus = Math.sin(_gam0 - _gam);
		double clam = Math.cos(_deltaLam);
		double slam = Math.sin(_deltaLam);
		double rjs = cg0*cgin + sg0*sgin*clam;
	
		double angle_rjs = Math.atan2(Math.sqrt(1-Math.pow(rjs,2)), rjs);
		double dg = -_gam0 + _gam + angle_rjs;
		double gamOut = _gam - dg;
	
		//_alp = asin(slam)*sgin)/sin(angle_rjs))
		//this expression for _alp only works sometimes, because
		//x and pi - x both have same sine 
	
		double  x = Math.sin(dg)*(cgin*sgin*(1-clam) + sg0minus*rjs) +
			Math.cos(dg)*(Math.pow(cgin,2) +Math.pow(sgin,2)*clam - cg0minus*rjs);
		double y = Math.sin(_gam0 -_gam + dg)*sgin*slam;
		_alp = Math.atan2(y,x);
		return gamOut;
	}


	/**
	 *writeStartControlledPhase writes exp(i _deltaLam |s><s|sign)
	 *where sign = (useMinus?-1:1)
	 */	
	public abstract void writeStartControlledPhase(boolean useMinus);
	
	/**
	 *writeTargetControlledPhase writes exp(i _alp |t><t|sign)
	 *where sign = (useMinus?-1:1)
	 */	
	public abstract void writeTargetControlledPhase(boolean useMinus);

}
