/**
 * This class stores parameters for embedding a circuit
 * within a larger circuit (one with more qubits)
 */
public class CktEmbedder {
	private int _numBitsBef;//number of bits in circuit before embedding
	private int _numBitsAft;//number of bits in circuit after embedding
	
	private int[]  _bitMap;//length=numBitsBef, bitAft[bitBef]
	private boolean[]  _isP0Control;//length=numBitsAft
	private boolean[]  _isP1Control;//length=numBitsAft
	private boolean[]  _isOldBit;//length=numBitsAft
	//_isP0Control and _isP1Control allow the new qubits 
	//to be P0 or P1 controls for the WHOLE circuit of old qubits
	//being embedded. The values of _isP0Control and _isP1Control
	//for the old qubits are ignored.

	/**
	 * 
	 */		
	public CktEmbedder(int numBits){
		_numBitsBef = numBits;
		_numBitsAft = numBits;
		_bitMap = null;
		_isP0Control = null;
		_isP1Control = null;
		_isOldBit = null;
	}
	
	/**
	 * 
	 */		
	public CktEmbedder(	int numBitsBef,
						int numBitsAft,
						int[] bitMap,
						boolean[] isP0Control,
						boolean[] isP1Control ){
		
		_numBitsBef = numBitsBef;
		_numBitsAft = numBitsAft;
		Throw.If(numBitsBef < 1, "embedding trivial circuit");
		Throw.If(numBitsBef > numBitsAft,"embedding into smaller number of bits");

		if(bitMap.length==numBitsBef){
			boolean[] isAlreadyPresent = Globals.F_vec(numBitsAft);				
			for(int bef=0; bef<numBitsBef; bef++){
				int aft = bitMap[bef];				
				if(aft<0 || aft>=numBitsAft || isAlreadyPresent[aft]){
					System.out.println("CktEmbedder's bitMap is not 1-1 " +
							"or is out of range");
					System.exit(0);		
				}else{
					isAlreadyPresent[aft] = true;
				}
			}
			_bitMap = bitMap;			
		}else{
			System.out.println("CktEmbedder's bitMap has wrong length");
			System.exit(0);
		}
		
		if(isP0Control.length==numBitsAft){
			_isP0Control = isP0Control;			
		}else{
			System.out.println("CktEmbedder's isP0Control has wrong length");
			System.exit(0);		
		}
		
		if(isP1Control.length==numBitsAft){
			_isP1Control = isP1Control;			
		}else{
			System.out.println("CktEmbedder's isP1Control has wrong length");
			System.exit(0);		
		}
		_isOldBit = Globals.F_vec(numBitsAft);
		for(int bef=0; bef<numBitsBef; bef++){
			_isOldBit[aft(bef)]  = true;
		}
	}
	/**
	 * 
	 */		
	public CktEmbedder(CktEmbedder emb){
		copy(emb);
	}
	/**
	 * 
	 */		
	public void copy(CktEmbedder emb){
		
		_numBitsBef = emb._numBitsBef;
		_numBitsAft = emb._numBitsAft;
		
		if(emb.isIdentity()){
			_bitMap=null;
			_isP0Control=null;
			_isP1Control=null;
			_isOldBit=null;
		}else{
		
			int j;
			
			_bitMap = new int[_numBitsBef];
			for(j=0; j<_numBitsBef; j++){
				_bitMap[j] = emb._bitMap[j];
			}
	
			_isP0Control = new boolean[_numBitsAft];
			for(j=0; j<_numBitsAft; j++){
				_isP0Control[j] = emb._isP0Control[j];
			}
			
			_isP1Control = new boolean[_numBitsAft];
			for(j=0; j<_numBitsAft; j++){
				_isP1Control[j] = emb._isP1Control[j];
			}
	
		
			_isOldBit = new boolean[_numBitsAft];
			for(j=0; j<_numBitsAft; j++){
				_isOldBit[j] = emb._isOldBit[j];
			}
		}

	}
	/**
	 * 
	 */	
	public int getNumBitsBef(){
		return _numBitsBef;
	}
	/**
	 * 
	 */	
	public int getNumBitsAft(){
		return _numBitsAft;
	}
	/**
	 * 
	 */	
	public int getNumNewBits(){
		return _numBitsAft - _numBitsBef;
	}
	/**
	 * 
	 */	
	public boolean isIdentity(){
		return _bitMap==null;
	}
	/**
	 * 
	 */	
	public int aft(int bef){
		if(!isIdentity()){
			return _bitMap[bef];
		}else{
			return  bef;
		}
	}
	/**
	 * 
	 */	
	public boolean isP0Control(int aft){
		if(!isIdentity()){
			return _isP0Control[aft];
		}else{
			return false;
		}
	}
	/**
	 * 
	 */	
	public boolean isP1Control(int aft){
		if(!isIdentity()){
			return _isP1Control[aft];
		}else{
			return false;
		}
	}
	/**
	 * 
	 */	
	public boolean isOldBit(int aft){
		if(!isIdentity()){
			return _isOldBit[aft];
		}else{
			return true;
		}
	}
	/**
	 * 
	 */	
	public boolean isNewBit(int aft){
		return !isOldBit(aft);
	}
	/**
	 * 
	 */	
	public int getNumNewControls(){
		int num=0;
		if(!isIdentity()){
			for(int aft=0; aft<_numBitsAft; aft++){
				if((!_isOldBit[aft]) && (_isP0Control[aft]||_isP1Control[aft])){
					num++;
				}
			}
		}
		return num;
	}
	/**
	 * 
	 */	
	public CktEmbedder getCompositeMapOfYouFollowedBy(CktEmbedder emb){
		if(this.isIdentity()) return  emb;
		if(emb.isIdentity()) return  this;
		Throw.If(this._numBitsAft != emb._numBitsBef, "can't chain embedders");
		int newNumBitsBef = this._numBitsBef;
		int newNumBitsAft = emb._numBitsAft;
		
		int[] newBitMap = new  int[newNumBitsBef];
		boolean[] newIsP0Control = new  boolean[newNumBitsAft];
		boolean[] newIsP1Control = new  boolean[newNumBitsAft];
		int  j;
		for(j=0;j<newNumBitsBef;j++){
			newBitMap[j]  = emb.aft(this.aft(j));
		}
		for(j=0;j<newNumBitsAft;j++){
			newIsP0Control[j] =  emb._isP0Control[j];
			newIsP1Control[j] =  emb._isP1Control[j];
		}
		for(j=0;j<this._numBitsAft;j++){
			newIsP0Control[emb.aft(j)] = this._isP0Control[j];
			newIsP1Control[emb.aft(j)] = this._isP1Control[j];
		}
		return  new CktEmbedder(newNumBitsBef,
							newNumBitsAft,
							newBitMap,
							newIsP0Control,
							newIsP1Control );
		
	}

}
