/**
 * This class writes English and Picture files for QOperAv.
 * QOperAv= Quantum Operator Average
 * The algorithm of QOperAv is described in ArXiv:0912.4402
 * "Use of Quantum Sampling to Calculate
 * Mean Values of Observables and Partition Function
 * of a Quantum System", by R.R. Tucci
 */


public abstract class QOperAvSEO_writer  extends SEO_writer{
	
	protected int  	_numAtomBits;//Atom is exp(iA \Delta t) in my paper
	protected int	_numProbeBits;//from Atom^1 to Atom^pow(2,numProbeBits-1)
	protected double	_gamma;//this is \gamma in my paper
	protected double	_dtOver2pi;//this is \Delta t /(2*PI) in my paper

	/**
	 * 
	 */	
	public QOperAvSEO_writer(
		String 	filePrefix,
		int 	numAtomBits,
	    int 	numProbeBits,
	    double 	gamma,
	    double 	dtOver2pi)	
	{
		super(filePrefix, 2 /*_numBits, temporary*/);
		//must reassign these variables because super constructor 
		//used  numBits=2
		_numBits = numAtomBits + numProbeBits + 1;
        _emb = new CktEmbedder(_numBits);
        
        _numAtomBits = numAtomBits;
		_numProbeBits = numProbeBits;
		
		Throw.If(gamma<=0, "Must have gamma > 0");
		_gamma = gamma;
		
		Throw.If(dtOver2pi<=0, "Must have dt/(2pi) > 0");
		_dtOver2pi = dtOver2pi;
		
		write();
		
        closeFiles();

	}
	/**
	 * 
	 */	
	private void write(){
		int j;
		
		//write atomic V
		//parameters for preEmb;
		int numBitsBef = _numAtomBits;
		int numBitsAft = _numBits;
		int[] bitMap = new  int[numBitsBef];
		for(j=0; j<numBitsBef; j++){
			bitMap[j] = j + _numProbeBits;
		}

		CktEmbedder savedEmb = new CktEmbedder(_emb);
		CktEmbedder preEmb = new  CktEmbedder(
					numBitsBef,
					numBitsAft,
					bitMap,
					Globals.F_vec(numBitsAft),//isP0Control
					Globals.F_vec(numBitsAft));//isP1Control
		_emb = preEmb.getCompositeMapOfYouFollowedBy(_emb);
		writeAtomicV();
		_emb  = savedEmb;//restoring _emb
	
		//write phase estimator
		//parameters for preEmb;
		numBitsBef = _numBits - 1;
		numBitsAft = _numBits;
		savedEmb = new CktEmbedder(_emb);
		preEmb = new  CktEmbedder(
					numBitsBef,
					numBitsAft,
					Globals.identityPermutation(numBitsBef),//bitMap
					Globals.F_vec(numBitsAft),//isP0Control
					Globals.F_vec(numBitsAft));//isP1Control
		_emb = preEmb.getCompositeMapOfYouFollowedBy(_emb);
		writePhaseEstimator();
		_emb  = savedEmb;//restoring _emb
		
		// write multiplexor
		int numMultiControls = _numProbeBits;
		int numAngles = 1<<numMultiControls;
		double[]  radAngles = new double[numAngles];
		for(j=0; j<numAngles; j++){
			double gammaF = _gamma*positiveFun(
					j/(_dtOver2pi*numAngles)
				);
			Throw.If(gammaF<0 || gammaF>1, 
					"Must have gamma*f() values in [0,1]");
			double cj = Math.sqrt(gammaF);
			double sj = Math.sqrt(1-gammaF);
			radAngles[j] = Math.atan2(sj, cj);
		}
		int targetBitPosition = _numBits - 1;
		
		MultiControls  controls = new  MultiControls(
				_numBits, numMultiControls);
		for(int c=0; c<numMultiControls; c++){
			controls.setControl(c, numMultiControls-c-1, numMultiControls-c-1);
		}

		writeControlledMultiplexorGate(
				targetBitPosition,
				controls,
				false, //wantHermitian
				new DoubleArray(radAngles));	
	}
	/**
	 * 
	 */	
	 protected abstract void  writeAtomicV();
	 /**
	 * 
	 */	
	 protected abstract void  writePhaseEstimator(); 
	 /**
	 * 
	 */
	protected abstract double positiveFun(double x);
}
