
/**
 * This class writes English and Picture files
 * for the full algorithm of Quibbs.
 * Quibbs generates a quantum circuit for doing Gibbs sampling 
 * of a classical Bayesian net.
 * (see ``Quibbs, a Code Generator
 * for Quantum Gibbs Sampling" by R.R. Tucci, arXiv:1004.2205 )
 */

public class QuibbsSEO_writer  extends SEO_writer{
	private CB_Net	_net;
	private int  	_numAtomBits;//Atom is general unitary transformation
	private int		_numProbeBits;//from Atom^1 to Atom^pow(2,numProbeBits-1)
	private int		_numPESteps;//Phase Estimation steps in stairs
	private int		_maxNumGroverSteps;
	private double	_gam0;//first gamma
	private double 	_gamTol; //gamma tolerance
	private double 	_deltaLam;//Lam=lambda
	private int[]	_startingNodeStates;
	private boolean[] 	_startingNitStates;
	private	boolean		_omitStairs;//use true only for debugging
	
	private int _numGroverSteps;
	private double _finGamma;

	/**
	 * 
	 */	
	public QuibbsSEO_writer(
			CB_Net 	net,//if this is null, will create _net 
	        String 	relPath,
			int[] 	startingNodeStates,//if this null, will set to random
									//if {0}, will set all components to zero	        
	    	int 	numProbeBits,
	    	int  	numPESteps,
	    	int 	maxNumGroverSteps,
			double	gam0,//If >=0, then this value of gam0 used
				//regardless of other stipulated initial values.
				//If <0, then this value of gam0 is ignored 
				//and one is calculated from other stipulated initial values.
			double 	gamTol,
			double 	deltaLam,
			boolean omitStairs)//only for debugging purposes									
	{ 	
	 	
    	super(relPath + "quibbs", 2/*dummy _numBits*/);
    	if(net==null){
    		net = new CB_Net(relPath + "states.txt", 
				relPath + "parents.txt",
				relPath + "probs.txt");
    	}
    	_net = net;

		int numNits = _net.getNumNits();
		//must reassign these variables because super constructor 
		//used  numBits=2
		_numBits = 2*numNits + numProbeBits*numPESteps;
        _emb = new CktEmbedder(_numBits);
		
		
		_numAtomBits = 2*numNits;
		_numProbeBits = numProbeBits;
		_numPESteps = numPESteps;
		_maxNumGroverSteps = maxNumGroverSteps;
		_gam0 = gam0;//will change later if gam0<0
		_gamTol = gamTol;
		_deltaLam = deltaLam;

		if(startingNodeStates==null){
			_startingNodeStates = _net.getRandomNodeStatesArray();
		}else if(startingNodeStates[0]==0 && startingNodeStates.length==1){
			_startingNodeStates = new int[_net.getNumNodes()];
			//created with all components 0
		}else{
			_startingNodeStates = startingNodeStates;
		}
		_net.checkNodeStatesArray(_startingNodeStates);

		_startingNitStates = 
			_net.getNitStatesFromNodeStates(_startingNodeStates);
		if(gam0<0){
			double  prob0 = _net.getProb(_startingNodeStates);
			_gam0 = 2*Math.acos(Math.sqrt(prob0));
		}
		_omitStairs = omitStairs;

		write();
		
        closeFiles();

	}
	/**
	 * 
	 */	
	public double getFinGamma(){
		return _finGamma;
	}
	/**
	 * 
	 */	
	public int getNumGroverSteps(){
		return _numGroverSteps;
	}
	/**
	 * 
	 */	
	private void write(){
		GroAdaWriter1 wr = new GroAdaWriter1(
				_maxNumGroverSteps,
				_gam0,
				_gamTol,
				_deltaLam);
		wr.write();
		_finGamma = wr.getGamma();
		_numGroverSteps = wr.getNumGroverSteps();
	}
	/**
	 * 
	 */	
	public int[] getStartingNodeStates(){
		return _startingNodeStates;
	}

	/**
	 * 
	 */	
    private class GroAdaWriter1 extends GroverAdaptiveWriter{
    	/**
	     * 
	     */
    	public GroAdaWriter1(
			int 	maxNumSteps,
			double	gam0,
			double 	gamTol,
			double 	deltaLam)
		{
			super(maxNumSteps, gam0, gamTol, deltaLam);
		}
	    /**
	     * 
	     */
    	public void writeStartControlledPhase(boolean useMinus){
    		//writes
    		//  exp(i*_alp 	I_probeBits)
    		//	   (		|0><0|_top half of atomBits)
    		//	   (		|s><s|_bottom half of atomBits)
    		//Replace i by -i if useMinus==true
    		int numNits = _numAtomBits/2;
    		int numNonPhaseControls = _numAtomBits-1;
    		//one bit has controlled phase
    		TFControls    controls =  
    			new TFControls(_numBits, numNonPhaseControls);
    		int bitPos = _numBits;
    		int ni=numNits;
    		for(int c=0; c<numNits; c++){
    			bitPos--;
    			ni--;
    			controls.setControl(c, bitPos, _startingNitStates[ni]);
    		}
    		for(int c=numNits; c<numNonPhaseControls; c++){
    			bitPos--;
    			controls.setControl(c, bitPos, false);
    		}
    		writeControlledOneBitGate(
    			_numPESteps*_numProbeBits,//targetBitPosition
		        controls,
		        new OneBitP0PhaseFactor(useMinus?-_alp:_alp));
    	}

    	/**
	     * 
	     */                        
    	public void writeTargetControlledPhase(boolean useMinus){
    		QuibbsStairsSEO_writer 	sw = new QuibbsStairsSEO_writer(
    		        _englishOut,
    		        _pictureOut,
    		        false,//doWrite
    		        _emb,
    				_numProbeBits,
    				new CB_NetQEmb(_net, true/*usingEffiMeth*/),
    				_numPESteps);
        		if(!_omitStairs){sw.write();}
        		//Dec 7, 2013: Maybe should have used writeHermitian() here
        		//and write() below. Should check someday
        		writeCoreOfTargetControlledPhase(useMinus);
        		if(!_omitStairs){sw.writeHermitian();}
        }
    	/**
    	 * 
    	 */	
    	private void writeCoreOfTargetControlledPhase(boolean  useMinus){
    		//writes
    		//  exp(i*_deltaLam	|0><0|_probeBits)
    		//	   (			I_atomBits      )
    		//Replace i by -i if useMinus==true
    		int  numNonPhaseControls = _numPESteps*_numProbeBits-1;
    		//one bit has controlled phase
    		TFControls    controls =  
    			new TFControls(_numBits, numNonPhaseControls);
    		int bitPos = numNonPhaseControls;
    		for(int c=0; c<numNonPhaseControls; c++){
    			bitPos--;
    			controls.setControl(c, bitPos, false);
    		}
    		writeControlledOneBitGate(
    			numNonPhaseControls,//targetBitPosition
		        controls,
		        new OneBitP0PhaseFactor(useMinus?-_deltaLam:_deltaLam));
    	}
    }
}

