import java.io.PrintWriter;


/**
 * This class writes English and Picture files
 * for the gamma qubit controlled portion in |s>.
 * (lonely H acting on gamma qubit is also included)
 * |s> is the AFGA starting state in the paper
 * ``Quantum Circuit for Discovering from Data the
 * Structure of Classical Bayesian Networks" by R.R. Tucci, arXiv:?
 */

public class JenGammaControlledSEO_writer  extends SEO_writer{
	private int _numNodes;//the number of nodes of the graph being discovered
	private int _maxNumPa;//maximum number of parents of the graph, 
						//is <= (numNodes-1)
	private int[] _nodeOwnerOfBit;
	/**
	 * 
	 */	
	public JenGammaControlledSEO_writer(
	        String 	filePrefix,
	        int numNodes,
	        int maxNumPa)
	{ 	 	
		super(filePrefix, 2 /*_numBits, temporary*/);
		Throw.If(maxNumPa>=numNodes, 
			"Must have (Maximum Number of Parents)<(Number of Nodes)" );
		_numNodes = numNodes;
		_maxNumPa = maxNumPa;
		_numBits = JenMain.calcNumBits(_numNodes, _maxNumPa);
		_nodeOwnerOfBit = new int[_numBits];
		fillOwnershipArray();
		
		_emb = new CktEmbedder(_numBits);
    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public JenGammaControlledSEO_writer(
		PrintWriter  englishOut,
		PrintWriter  pictureOut,
		Boolean doWrite,
		CktEmbedder emb,
        int numNodes,
        int maxNumPa)
    {
    	super(englishOut, pictureOut, emb);
		Throw.If(maxNumPa>=numNodes, 
				"Must have (Maximum Number of Parents)<(Number of Nodes)" );
		_numNodes = numNodes;
		_maxNumPa = maxNumPa;
		_numBits = JenMain.calcNumBits(_numNodes, _maxNumPa);
		_nodeOwnerOfBit = new int[_numBits];
		fillOwnershipArray();
 	
    	if(doWrite){write();}
    }
	/**
	 * Every qubit is owned by one of the nodes 
	 * except the gamma and omega which are owned by -1
	 */
    private int getNodeOwnerOfBit(int bit){
    	int owner=-1;
    	if(bit<_numNodes){
    		owner = bit;
    	}else{
    		int relPos = bit - _numNodes;//relative position of bit
    		int combLen, numCombs;
    		combLen=0;
    		numCombs=1;
    		while(combLen<=_maxNumPa){
    			numCombs = Combinations.get_nCr(_numNodes-1, combLen); 
    			if(relPos>=_numNodes*numCombs){
    				relPos -=_numNodes*numCombs;
    			}else{
    				break;
    			}
    			combLen++;
    		}
			owner = relPos/numCombs;
    	}
    	return owner;
    }
	/**
	 * 
	 */
    private void fillOwnershipArray(){
		for(int bit=0; bit<_numBits; bit++){
			_nodeOwnerOfBit[bit] = getNodeOwnerOfBit(bit);
		}    	
    }
	/**
	 * 
	 */
	public void write(){
		int gammaBit = _numBits-3;
		
		writeOneBitGate(gammaBit, new OneBitHadamard());

		int combLen;// this refers to the length of the combination set, set of parents
		int[][] combs;
		int numCombs, node, com;
		double angle;
		int numTrols, j, bitPos;		
		Jen_hFun hFun = new Jen_hFun(_numNodes);
		int gateCounter=0;
		for(combLen=0; combLen<=_maxNumPa; combLen++){
			if(combLen==0){
				combs = null;
				numCombs=1;
			}else{
				combs = Combinations.getCombinations(combLen, _numNodes-1);
				//the array combs is numCombs=nCr times combLen
				numCombs = combs.length;
			}
			for(node=0; node<_numNodes; node++){
				for(com=0; com<numCombs; com++){
					angle = Math.asin(hFun.h(node, combs==null?null:combs[com]));
					j = 2;// two true controls, the rest are false ones
					for(bitPos=gateCounter + _numNodes -1; bitPos>=_numNodes; bitPos--){
						if(_nodeOwnerOfBit[bitPos]==node){
							j++;
						}						
					}
					numTrols = j + combLen;
					TFControls trols = new TFControls(_numBits, numTrols);
					trols.setControl(0, gammaBit/*bitPos*/, true/*type*/);
					trols.setControl(1, gateCounter + _numNodes/*bitPos*/, true/*type*/);
					j=1;
					for(bitPos=gateCounter + _numNodes -1; bitPos>=_numNodes; bitPos--){
						if(_nodeOwnerOfBit[bitPos]==node){
							j++;
							trols.setControl(j, bitPos/*bitPos*/, false/*type*/);
						}						
					}
					if(combLen>0){
						for(int combIndex=0; combIndex<combLen; combIndex++){	
							if(combs[com][combIndex]==node){
								j++;
								trols.setControl(j, _numNodes-1/*bitPos*/, true/*type*/);
								break;		
							}
						}
						for(bitPos=_numNodes-2; bitPos>=0; bitPos--){
							for(int combIndex=0; combIndex<combLen; combIndex++){	
								if(bitPos!=node && combs[com][combIndex]==bitPos){
									j++;
									trols.setControl(j, bitPos/*bitPos*/, true/*type*/);		
									break;
								} 
							}	
						}
					}
					gateCounter++;
					trols.setControl(0, gammaBit/*bitPos*/, true/*type*/);
					writeControlledOneBitGate(
						node, //targetBitPosition, 
						trols, 
						new OneBitRot(-angle, OneBitRot.Y_AXIS));
					trols.setControl(0, gammaBit/*bitPos*/, false/*type*/);
					writeControlledOneBitGate(
							node, //targetBitPosition, 
							trols, 
							new OneBitHadamard());
				}
			}	
		}
	}
	/**
	 * writes Hermitian Conjugate
	 */
	public void writeHermitian(){
		int gammaBit = _numBits-3;
		
		int combLen;// this refers to the length of the combination set, set of parents
		int[][] combs;
		int numCombs, node, com;
		double angle;
		int numTrols, j, bitPos;
		Jen_hFun hFun = new Jen_hFun(_numNodes);
		int gateCounter=0;//decreases in writeHermitian() but grows in write()
		for(combLen=_maxNumPa; combLen>=0; combLen--){
			if(combLen==0){
				combs = null;
				numCombs=1;
			}else{
				combs = Combinations.getCombinations(combLen, _numNodes-1);
				//the array combs is numCombs=nCr times combLen
				numCombs = combs.length;
			}
			for(node=_numNodes-1; node>=0; node--){
				for(com=numCombs-1; com>=0; com--){
					angle = Math.asin(hFun.h(node, combs==null?null:combs[com]));
					j = 2;// two true controls, the rest are false ones
					for(bitPos=gateCounter + _numBits -5; bitPos>=_numNodes; bitPos--){
						if(_nodeOwnerOfBit[bitPos]==node){
							j++;
						}						
					}
					numTrols = j + combLen;
					TFControls trols = new TFControls(_numBits, numTrols);
					trols.setControl(0, gammaBit/*bitPos*/, true/*type*/);
					trols.setControl(1, gateCounter + _numBits - 4/*bitPos*/, true/*type*/);
					j=1;
					for(bitPos=gateCounter + _numBits -5; bitPos>=_numNodes; bitPos--){
						if(_nodeOwnerOfBit[bitPos]==node){
							j++;
							trols.setControl(j, bitPos/*bitPos*/, false/*type*/);
						}						
					}
					if(combLen>0){
						for(int combIndex=0; combIndex<combLen; combIndex++){	
							if(combs[com][combIndex]==node){
								j++;
								trols.setControl(j, _numNodes-1/*bitPos*/, true/*type*/);
								break;		
							}
						}
						for(bitPos=_numNodes-2; bitPos>=0; bitPos--){
							for(int combIndex=0; combIndex<combLen; combIndex++){	
								if(bitPos!=node && combs[com][combIndex]==bitPos){
									j++;
									trols.setControl(j, bitPos/*bitPos*/, true/*type*/);		
									break;
								} 
							}	
						}
					}
					gateCounter--;
					trols.setControl(0, gammaBit/*bitPos*/, false/*type*/);
					writeControlledOneBitGate(
							node, //targetBitPosition, 
							trols, 
							new OneBitHadamard());
					trols.setControl(0, gammaBit/*bitPos*/, true/*type*/);
					writeControlledOneBitGate(
						node, //targetBitPosition, 
						trols, 
						new OneBitRot(angle, OneBitRot.Y_AXIS));
				}
			}	
		}
		writeOneBitGate(gammaBit, new OneBitHadamard());

	}

	/**
	 * 
	 */
    public static void main(String[] args){

    	int numNodes = 3;
    	int maxNumPa = 2;
        int numBits = JenMain.calcNumBits(numNodes, maxNumPa);
  
    	String filePrefix = "gam-con";

    	JenGammaControlledSEO_writer  w = new JenGammaControlledSEO_writer(
    			openEngFile(filePrefix),
    			openPicFile(filePrefix),
    			true, //doWrite
    			new CktEmbedder(numBits),        
    			numNodes,
    			maxNumPa);

    	w.writeHermitian();
    	w.closeFiles();
    }
}

