
/**
 * This class writes English and Picture files
 * for the full algorithm of qMobius.
 * qMobius generates a quantum circuit for calculating
 * Mobius transform of a function using the method described in the paper
 * ``Quantum Circuit for Calculating Mobius-like Transforms
 * Via Grover-like algorithm" by R.R. Tucci, arXiv:?
 */


public class MobSEO_writer  extends SEO_writer{
	private int _numPsiBits;// the number of qubits in psi
	private Boolean[] _cVec;//dimension is numPsiBits
		//_cVec[j]=c_j in my paper mentioned in header of this file
	double _estP1overP0;//an estimate of |z_1|^2/|z_0|^2
						//where P0=|z0|^2 and P1= |z_1|^2, 
						//In general, P0+P1 != 1
	double _P0;
	private int		_maxNumGroverSteps;
	private double	_gam0;//first gamma
	private double 	_gamTol; //gamma tolerance
	private double 	_deltaLam;//Lam=lambda
	
	private int _numGroverSteps;
	private double _finGamma;
	/**
	 * 
	 */	
	public MobSEO_writer(
	        String 	filePrefix,	        
	        Boolean[] cVec,
	        double 	estP1overP0,
	    	int 	maxNumGroverSteps,
			double	gam0,//If >=0, then this value of gam0 used
				//regardless of other stipulated initial values.
				//If <0, then this value of gam0 is ignored 
				//and one is calculated from other stipulated initial values.
			double 	gamTol,
			double 	deltaLam)									
	{ 	
	 	
		super(filePrefix, 2/*dummy _numBits*/);
		//must reassign these variables because super constructor 
		//used  numBits=2
		_numPsiBits = cVec.length;
		_numBits = MobMain.calcNumBits(_numPsiBits);
		_cVec = new Boolean[_numPsiBits];
		for(int j=0; j<_numPsiBits; j++){
			_cVec[j] = cVec[j];
		}
		_estP1overP0 = estP1overP0;
		setP0();//must come after _numNodes and _maxNumPa are set
        _emb = new CktEmbedder(_numBits);
		
		

		_maxNumGroverSteps = maxNumGroverSteps;
		_gam0 = gam0;//will change later if gam0<0
		_gamTol = gamTol;
		_deltaLam = deltaLam;

		if(gam0<0){
			//prob_st = p = <s|t>^2 = |z_0|^2 + |z_1|^2
			//P0=|z_0|^2, P1=|z_1|^2
			double  prob_st = _P0*(1 + estP1overP0);
			Throw.If(estP1overP0<0 || prob_st>1,
				"The estimate of |z_1|^2 / |z_0|^2 must be > 0 and < "
				+ (1/_P0 - 1));
			_gam0 = 2*Math.acos(Math.sqrt(prob_st));
		}

		write();
		
        closeFiles();

	}
	/**
	 * 
	 */	
	public double getFinGamma(){
		return _finGamma;
	}
	/**
	 * 
	 */	
	public int getNumGroverSteps(){
		return _numGroverSteps;
	}
	/**
	 * calculates P0= |z_0|^2 = 1/2^{numPsiBits + 1}
	 */
	public void setP0(){
		_P0= 1.0/(1<<(_cVec.length + 1));
	}	
	/**
	 * 
	 */	
	double getP0(){
		return _P0;
	}
	/**
	 * 
	 */	
	private void write(){
		new MobStartEvOpSEO_writer(
		        _englishOut,
		        _pictureOut,
    			true, //doWrite
    			new CktEmbedder(_numBits),        
    			_cVec);
		GroAdaWriter1 wr = new GroAdaWriter1(
				_maxNumGroverSteps,
				_gam0,
				_gamTol,
				_deltaLam);
		wr.write();
		_finGamma = wr.getGamma();
		_numGroverSteps = wr.getNumGroverSteps();
	}


	/**
	 * 
	 */	
    private class GroAdaWriter1 extends GroverAdaptiveWriter{
    	/**
	     * 
	     */
    	public GroAdaWriter1(
			int 	maxNumSteps,
			double	gam0,
			double 	gamTol,
			double 	deltaLam)
		{
			super(maxNumSteps, gam0, gamTol, deltaLam);
		}
	    /**
	     * 
	     */
    	public void writeStartControlledPhase(boolean useMinus){
    		//writes
    		//exp(i*_alp |s><s|)
    		//Replace i by -i if useMinus==true
    		
    		//all qubits are controls except gamma qubit 
    		int numTrols = _numBits-1;
    		TFControls    controls =  
    			new TFControls(_numBits, numTrols);
    		controls.setControl(0, _numBits-1, false);//omega  qubit
    		int bitPos = _numBits-3;
    		for(int c=1; c<numTrols; c++){
    			controls.setControl(c, bitPos, false);
    			bitPos--;
    		}
    		MobStartEvOpSEO_writer w = new MobStartEvOpSEO_writer(
    		        _englishOut,
    		        _pictureOut,
        			false, //doWrite
        			_emb,        
        			_cVec);
    		
    		w.writeHermitian();
    		writeControlledOneBitGate(
        			_numBits-2,//targetBitPosition=gamma qubit
    		        controls,
    		        new OneBitP0PhaseFactor(useMinus?-_alp:_alp));
    		w.write();
    		
    	}

    	/**
	     * 
	     */                        
    	public void writeTargetControlledPhase(boolean useMinus){
    		//writes 
    		//exp(i*_deltaLam	|0><0|_omega  qubit)
    		//Replace i by -i if useMinus==true
   		
    		writeOneBitGate(
    			_numBits-1,//omega  qubit
    			new OneBitP0PhaseFactor(useMinus?-_deltaLam:_deltaLam));

        }
    }
}

