import java.io.*;


/**
 * This class writes English and Picture files
 * for: (1) the controlled gate that targets the omega qubit in |s>
 * (lonely sig_x acting on omega  qubit is also included), 
 * and (2) The cnot that targets the mu_0 qubit.
 * |s> is the AFGA starting state for Method A in the paper
 * ``Quantum Circuit for Calculating Symmetrized Functions
 * Via Grover-like algorithm" by R.R. Tucci, arXiv:?
 */

public class SymTargetingSEO_writer  extends SEO_writer{
	private int _numQudits;// the number of qudits in alpha
	private int _bitsPerDit;// the number of qubits in each qudit
	private Boolean[][] _cMat;//dimension is numQudits X bitsPerDit

	/**
	 * 
	 */	
	public SymTargetingSEO_writer(
	        String 	filePrefix,
	        Boolean[][] cMat)
	{ 	 	
		super(filePrefix, 2 /*_numBits, temporary*/);
		_numQudits = cMat.length;
		_bitsPerDit = cMat[0].length;
		_numBits = SymMain.calcNumBits(_numQudits, _bitsPerDit);
		_cMat = new Boolean[_numQudits][_bitsPerDit];
		for(int j=0; j<_numQudits; j++){
			for(int k=0; k<_bitsPerDit; k++){
				_cMat[j][k] = cMat[j][k];
			}
		}
		
		_emb = new CktEmbedder(_numBits);
    	write();

        closeFiles();

	}
	/**
	 * 
	 */
    public SymTargetingSEO_writer(
		PrintWriter  englishOut,
		PrintWriter  pictureOut,
		Boolean doWrite,
		CktEmbedder emb,
        Boolean[][] cMat)
    {
    	super(englishOut, pictureOut, emb);
		_numQudits = cMat.length;
		_bitsPerDit = cMat[0].length;
		_numBits = SymMain.calcNumBits(_numQudits, _bitsPerDit);
		_cMat = new Boolean[_numQudits][_bitsPerDit];
		for(int j=0; j<_numQudits; j++){
			for(int k=0; k<_bitsPerDit; k++){
				_cMat[j][k] = cMat[j][k];
			}
		}
  	
    	if(doWrite){write();}
    }
	/**
	 * 
	 */
	private void writeOmegaGate(){
		
		int numTrols = _numBits -3;
		int numVbits = numTrols - _numQudits*_bitsPerDit;
		TFControls trols = new TFControls(_numBits, numTrols);
		
		int trolPos = _numBits -4;
		int c =0;
		while(c<numVbits){
			trols.setControl(c, trolPos, false);
			trolPos--;
			c++;
		}
		for(int j=_numQudits-1; j>=0; j--){
			for(int k=_bitsPerDit-1; k>=0; k--){
				trols.setControl(c, trolPos, _cMat[j][k]);
				c++;
				trolPos--;
			}
		}
		writeControlledOneBitGate(
				_numBits - 1, //targetBitPosition, 
				trols, 
				new SigX());
	}
	/**
	 * 
	 */
	public void write(){
		writeOneBitGate(_numBits -1, new SigX());
		TFControls trols = new TFControls(_numBits, 1);
		trols.setControl(0, _numBits - 3, true);
		writeControlledOneBitGate(
				_numBits - 2, //targetBitPosition, mu_0
				trols, 
				new SigX());
		writeOmegaGate();
	}
	/**
	 * writes Hermitian Conjugate
	 */
	public void writeHermitian(){
		writeOmegaGate();
		TFControls trols = new TFControls(_numBits, 1);
		trols.setControl(0, _numBits - 3, true);
		writeControlledOneBitGate(
				_numBits - 2, //targetBitPosition, mu_0
				trols, 
				new SigX());
		writeOneBitGate(_numBits -1, new SigX());
	}

	/**
	 * 
	 */
    public static void main(String[] args){

    	Boolean[][] cMat={
    			{true, false, true},
    			{true, true, false},
    			{false, false, false},
    			{false, true, true}			
    	};
    	
    	int numQudits = cMat.length; 
    	int bitsPerDit = cMat[0].length;  	
        int numBits = SymMain.calcNumBits(numQudits, bitsPerDit);
  
    	String filePrefix = "targ";
    	
    	SymTargetingSEO_writer  w = new SymTargetingSEO_writer(
    			openEngFile(filePrefix),
    			openPicFile(filePrefix),
    			true, //doWrite
    			new CktEmbedder(numBits),        
    			cMat);

    	w.writeHermitian();
    	w.closeFiles();
    	
		for(int j=numQudits-1; j>=0; j--){
			for(int k=bitsPerDit-1; k>=0; k--){
				System.out.printf("%d",cMat[j][k]==true?1:0);
			}
		}

    }
}

